from copy import deepcopy
from typing import Callable
from ferrmion import FermionHamiltonian
import numpy as np
import pytest
from ferrmion.encode.ternary_tree import (
    TernaryTree,
    TTNode,
    JW,
    JordanWigner,
    BK,
    BravyiKitaev,
    JKMN,
    ParityEncoding,
    PE,
    TTFlatpack,
)
from ferrmion.utils import symplectic_hash, symplectic_unhash, symplectic_to_pauli
from ferrmion.hamiltonians import molecular_hamiltonian
from ferrmion.core import standard_symplectic_matrix, flatpack_symplectic_matrix
from .conftest import diagonalise_pauli_hamiltonian
import hypothesis
from hypothesis import given,strategies as st
from hypothesis.extra.numpy import arrays
import logging
logger = logging.getLogger(__name__)

try:
    import symmer
    from symmer import PauliwordOp, QuantumState
except ImportError:
    logger.warning("Could not import symmer.")
    symmer = None

@pytest.fixture
def six_mode_tree():
    return TernaryTree(n_modes=6, root_node=TTNode())


@pytest.fixture(scope="module")
def bonsai_paper_tree():
    tt = TernaryTree(n_modes=11)
    tt = tt.add_node("x")
    tt = tt.add_node("y")
    tt = tt.add_node("z")
    tt = tt.add_node("xx")
    tt = tt.add_node("xy")
    tt = tt.add_node("yx")
    tt = tt.add_node("yy")
    tt = tt.add_node("yz")
    tt = tt.add_node("zz")
    tt = tt.add_node("yzz")
    tt.enumeration_scheme = tt.default_enumeration_scheme()
    return tt


def test_standard_encoding_functions(six_mode_tree):
    # Test function aliases
    assert JW(6) == JordanWigner(6)
    assert BK(6) == BravyiKitaev(6)

    # Test TT function aliases
    assert six_mode_tree.JW() == JW(6)
    assert six_mode_tree.JordanWigner() == JordanWigner(6)
    assert six_mode_tree.BK() == BK(6)
    assert six_mode_tree.BravyiKitaev() == BravyiKitaev(6)
    assert six_mode_tree.JKMN() == JKMN(6)
    assert six_mode_tree.ParityEncoding() == ParityEncoding(6)

    # Test inequality by type
    assert JW(6) != BK(6)
    assert JW(6) != JKMN(6)
    assert JW(6) != ParityEncoding(6)
    assert BK(6) != JKMN(6)
    assert BK(6) != ParityEncoding(6)
    assert JKMN(6) != ParityEncoding(6)

    # Test inequality
    assert JW(6) != JW(5)
    assert JW(6) != JW
    assert JW(6) != "JW(6)"

    jw_different_enumeration = JW(6)
    jw_different_enumeration.enumeration_scheme["z"] = JW(6).enumeration_scheme["zz"]
    jw_different_enumeration.enumeration_scheme["zz"] = JW(6).enumeration_scheme["z"]
    assert JW(6) != jw_different_enumeration


def test_default_enumeration_scheme(six_mode_tree):
    assert six_mode_tree.default_enumeration_scheme() == {"": (0, 0)}
    jkmn = six_mode_tree.JKMN()
    assert jkmn.default_enumeration_scheme() == {
        "": (0, 0),
        "x": (1, 1),
        "y": (2, 2),
        "z": (3, 3),
        "xx": (4, 4),
        "xy": (5, 5),
    }


def test_invalid_enumeration_scheme(six_mode_tree):
    jkmn = six_mode_tree.JKMN()
    # Not enough qubit labels
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {
            "": (0, 0),
            "x": (1, 1),
            "y": (2, 2),
            "z": (3, 3),
            "xx": (4, 4),
            "xy": (5, 4),
        }
    assert "Expected 6 qubit labels" in str(exc.value)

    # Not enough mode labels
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {
            "": (0, 0),
            "x": (1, 1),
            "y": (2, 2),
            "z": (3, 3),
            "xx": (5, 4),
            "xy": (5, 5),
        }
    assert "Invalid mode labels" in str(exc.value)

    # Mode label not in range
    with pytest.raises(ValueError) as exc:
        jkmn.enumeration_scheme = {
            "": (6, 0),
            "x": (1, 1),
            "y": (2, 2),
            "z": (3, 3),
            "xx": (4, 4),
            "xy": (5, 5),
        }
    assert "Invalid mode labels" in str(exc.value)


def test_valid_enumeration_scheme(six_mode_tree):
    jkmn = six_mode_tree.JKMN()
    # We allow any qubit labels
    jkmn.enumeration_scheme = {
        "": (3, 10),
        "x": (2, 50),
        "y": (0, 30),
        "z": (1, 40),
        "xx": (4, 20),
        "xy": (5, 0),
    }


    jkmn.enumeration_scheme = {
        "": (3, 1),
        "x": (2, 5),
        "y": (0, 3),
        "z": (1, 4),
        "xx": (4, 2),
        "xy": (5, 0),
    }


def test_bravyi_kitaev(six_mode_tree):
    tt = six_mode_tree.BK()
    assert tt.root_node.branch_strings == {
        "xxzy",
        "xxzx",
        "xxzz",
        "xzx",
        "xzy",
        "xzz",
        "xxy",
        "xxxx",
        "y",
        "xy",
        "z",
        "xxxz",
        "xxxy",
    }

    assert tt.root_node.child_strings == ["", "x", "xx", "xz", "xxx", "xxz"]

    assert tt.as_dict() == {
        "x": {
            "x": {
                "x": {"x": None, "y": None, "z": None},
                "y": None,
                "z": {"x": None, "y": None, "z": None},
            },
            "y": None,
            "z": {"x": None, "y": None, "z": None},
        },
        "y": None,
        "z": None,
    }

    assert tt.default_enumeration_scheme() == {
        "": (0, 0),
        "x": (1, 1),
        "xx": (2, 2),
        "xz": (3, 3),
        "xxx": (4, 4),
        "xxz": (5, 5),
    }

    assert tt.string_pairs == {
        "": ("xzz", "y"),
        "x": ("xxzz", "xy"),
        "xx": ("xxxz", "xxy"),
        "xz": ("xzx", "xzy"),
        "xxx": ("xxxx", "xxxy"),
        "xxz": ("xxzx", "xxzy"),
    }

    assert tt.branch_pauli_map == {
        "xxzy": "XXZIIY",
        "xxzx": "XXZIIX",
        "xxzz": "XXZIIZ",
        "xzx": "XZIXII",
        "xzy": "XZIYII",
        "xzz": "XZIZII",
        "xxy": "XXYIII",
        "xxxx": "XXXIXI",
        "y": "YIIIII",
        "xy": "XYIIII",
        "xxxz": "XXXIZI",
        "xxxy": "XXXIYI",
        "z": "ZIIIII",
    }

    assert tt.n_qubits == len(tt.root_node.child_strings)
    assert np.all(
        tt._build_symplectic_matrix()[1]
        == np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0],
                [1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1],
                [1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0],
                [1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0],
                [1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1],
            ],
            dtype=np.int8,
        )
    )

    for line in tt._build_symplectic_matrix()[1]:
        assert np.all(line == symplectic_unhash(symplectic_hash(line), len(line)))


def tests_bonsai_paper_tree(bonsai_paper_tree):
    tt = bonsai_paper_tree
    assert tt.root_node.branch_strings == {
        "xyz",
        "zzy",
        "yyx",
        "yxz",
        "yzx",
        "yyy",
        "yzzx",
        "xyx",
        "xxx",
        "xxz",
        "yxx",
        "yzy",
        "xyy",
        "xxy",
        "yzzz",
        "yyz",
        "yxy",
        "zx",
        "zzz",
        "xz",
        "yzzy",
        "zzx",
        "zy",
    }

    assert tt.root_node.child_strings == [
        "",
        "x",
        "y",
        "z",
        "xx",
        "xy",
        "yx",
        "yy",
        "yz",
        "zz",
        "yzz",
    ]

    assert tt.as_dict() == {
        "x": {
            "x": {"x": None, "y": None, "z": None},
            "y": {"x": None, "y": None, "z": None},
            "z": None,
        },
        "y": {
            "x": {"x": None, "y": None, "z": None},
            "y": {"x": None, "y": None, "z": None},
            "z": {"x": None, "y": None, "z": {"x": None, "y": None, "z": None}},
        },
        "z": {"x": None, "y": None, "z": {"x": None, "y": None, "z": None}},
    }

    assert tt.default_enumeration_scheme() == {
        "": (0, 0),
        "x": (1, 1),
        "y": (2, 2),
        "z": (3, 3),
        "xx": (4, 4),
        "xy": (5, 5),
        "yx": (6, 6),
        "yy": (7, 7),
        "yz": (8, 8),
        "zz": (9, 9),
        "yzz": (10, 10),
    }

    assert tt.string_pairs == {
        "": ("xz", "yzzz"),
        "x": ("xxz", "xyz"),
        "y": ("yyz", "yxz"),
        "z": ("zx", "zy"),
        "xx": ("xxx", "xxy"),
        "xy": ("xyy", "xyx"),
        "yx": ("yxy", "yxx"),
        "yy": ("yyx", "yyy"),
        "yz": ("yzy", "yzx"),
        "zz": ("zzx", "zzy"),
        "yzz": ("yzzy", "yzzx"),
    }

    assert tt.branch_pauli_map == {
        "xyz": "XYIIIZIIIII",
        "zzy": "ZIIZIIIIIYI",
        "yyx": "YIYIIIIXIII",
        "yxz": "YIXIIIZIIII",
        "yzx": "YIZIIIIIXII",
        "yyy": "YIYIIIIYIII",
        "yzzx": "YIZIIIIIZIX",
        "xyx": "XYIIIXIIIII",
        "xxx": "XXIIXIIIIII",
        "xxz": "XXIIZIIIIII",
        "yxx": "YIXIIIXIIII",
        "yzy": "YIZIIIIIYII",
        "xyy": "XYIIIYIIIII",
        "xxy": "XXIIYIIIIII",
        "yzzz": "YIZIIIIIZIZ",
        "yyz": "YIYIIIIZIII",
        "yxy": "YIXIIIYIIII",
        "zx": "ZIIXIIIIIII",
        "xz": "XZIIIIIIIII",
        "yzzy": "YIZIIIIIZIY",
        "zzx": "ZIIZIIIIIXI",
        "zy": "ZIIYIIIIIII",
        "zzz": "ZIIZIIIIIZI",
    }

    assert tt.n_qubits == len(tt.root_node.child_strings)
    assert np.all(
        tt._build_symplectic_matrix()[1]
        == np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1],
                [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1],
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0],
            ],
            dtype=np.int8,
        )
    )

    for line in tt._build_symplectic_matrix()[1]:
        assert np.all(line == symplectic_unhash(symplectic_hash(line), len(line)))

def test_default_mode_op_map(water_tt):
    assert np.all(water_tt.default_mode_op_map == [*range(water_tt.n_qubits)])

@pytest.mark.parametrize("n_modes", [1,5,10,20])
@pytest.mark.parametrize("encoding,name", [(JordanWigner, "JW"), (ParityEncoding, "PE"), (BravyiKitaev, "BK"), (JKMN, "JKMN")])
def test_core_standard_encodings(n_modes,encoding,name):
    n_modes = 20
    i,s = encoding(20)._build_symplectic_matrix()
    ci, cs = standard_symplectic_matrix(name,20)
    assert np.all(i==ci)
    assert np.all(s==cs)

@pytest.mark.parametrize("optimisation", ["naive", "anneal", "topphatt"])
@pytest.mark.parametrize("encoding", [JW, BK, ParityEncoding, JKMN])
def test_encode_num_terms_equal_expected(encoding: Callable[[int], TernaryTree], optimisation:str, mol_data_sets: dict):
    ones = mol_data_sets["ones"]
    twos = mol_data_sets["twos"]
    e_nuc = mol_data_sets["constant_energy"]
    n_modes = ones.shape[0]
    fham = FermionHamiltonian(terms = {"+-":ones,"++--":twos}, constant_energy=e_nuc)
    initial_ones = deepcopy(ones)
    initial_twos = deepcopy(twos)

    match optimisation:
        case "naive":
            qham = encoding(fham.n_modes).encode(fham)
        case "anneal":
            qham = encoding(fham.n_modes).encode_annealed(fham)
        case "topphatt":
            qham = encoding(fham.n_modes).encode_topphatt(fham)

    assert len(qham) == mol_data_sets["num_terms"]

@pytest.mark.parametrize("optimisation", ["naive", "anneal", "topphatt"])
@pytest.mark.parametrize("encoding", [JW, BK, ParityEncoding, JKMN])
def test_encode_h2_eigvals_equal_expected(encoding: Callable[[int], TernaryTree], optimisation:str, h2_mol_data_sets: dict):
    ones = h2_mol_data_sets["ones"]
    twos = h2_mol_data_sets["twos"]
    e_nuc = h2_mol_data_sets["constant_energy"]
    n_modes = ones.shape[0]
    fham = FermionHamiltonian(terms = {"+-":ones,"++--":twos}, constant_energy=e_nuc)
    initial_ones = deepcopy(ones)
    initial_twos = deepcopy(twos)

    match optimisation:
        case "naive":
            qham = encoding(fham.n_modes).encode(fham)
        case "anneal":
            qham = encoding(fham.n_modes).encode_annealed(fham)
        case "topphatt":
            qham = encoding(fham.n_modes).encode_topphatt(fham)
    diag  = diagonalise_pauli_hamiltonian(qham, 2*n_modes)

    assert np.all(initial_ones == ones)
    assert np.all(initial_twos == twos)
    assert np.allclose(np.sort(diag), np.sort(h2_mol_data_sets["eigvals"]))

@pytest.mark.parametrize("optimisation", ["naive", "topphatt"])
@pytest.mark.parametrize("encoding", [JW])
def test_encode_jw_water_eigvals_equal_expected(encoding: Callable[[int], TernaryTree], optimisation:str,  water_data: dict):
    ones = water_data["ones"]
    twos = water_data["twos"]
    e_nuc = water_data["constant_energy"]
    n_modes = ones.shape[0]

    fham = FermionHamiltonian(terms = {"+-":ones,"++--":twos}, constant_energy=e_nuc)

    match optimisation:
        case "naive":
            qham = encoding(fham.n_modes).encode(fham)
        # Takes too long for tests!
        # case "anneal":
            # qham = encoding(fham.n_modes).encode_annealed(fham)
        case "topphatt":
            qham = encoding(fham.n_modes).encode_topphatt(fham)
    assert np.isclose(qham["I"*14], -46.465600781952176)
    diag = diagonalise_pauli_hamiltonian(qham, 2)

    assert np.allclose(np.sort(diag), np.sort(water_data["eigvals"])[:2])

@given(arrays(dtype=np.bool, shape=st.integers(1, 9)))
def test_naive_jw_hf_state_unchanged(fermionic_hf_state):
    tree = JordanWigner(len(fermionic_hf_state))
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    print(f"fermionic HF {fermionic_hf_state}")
    qubit_hf_state = tree.hartree_fock_state(
        fermionic_hf_state=fermionic_hf_state,
        mode_op_map=[*range(len(fermionic_hf_state))],
    )
    assert np.all(qubit_hf_state == fermionic_hf_state)

@given(mode_op_map=st.permutations([*range(10)]), n_electrons=st.integers(min_value=1, max_value=10))
def test_enumerated_jw_hf_state_match_reordered_naive(mode_op_map, n_electrons):
    fermionic_hf_state = np.array([True] * n_electrons + [False] * (10-n_electrons), dtype=np.bool)

    tree = JordanWigner(len(fermionic_hf_state))
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    print(f"\nfermionic HF {fermionic_hf_state}")
    print(f"Enumeration {mode_op_map}")
    naive_qubit_hf_state = tree.hartree_fock_state(
        fermionic_hf_state=fermionic_hf_state,
        mode_op_map=[*range(len(fermionic_hf_state))],
    )

    print(f"naive {naive_qubit_hf_state}")
    enumerated_qubit_hf_state = tree.hartree_fock_state(
        fermionic_hf_state=fermionic_hf_state,
        mode_op_map=mode_op_map,
    )
    print(f"enumerated {enumerated_qubit_hf_state}")
    expected_emnumerated = np.array([False] * 10, dtype=np.bool)
    expected_emnumerated[mode_op_map[:n_electrons]] = True
    print(f"expected {enumerated_qubit_hf_state}")

    assert np.all(naive_qubit_hf_state == fermionic_hf_state)
    assert np.all(enumerated_qubit_hf_state == expected_emnumerated)

@pytest.mark.skipif(symmer is None, reason="Dependency group test not installed.")
@pytest.mark.parametrize("encoding", [JW, PE, BK])
def test_naive_water_hf_energy_correct(encoding, water_data):
    fermionic_hf_state = np.array([True]*10 + [False] * 4, dtype=np.bool)

    tree = encoding(len(fermionic_hf_state))
    tree.enumeration_scheme = tree.default_enumeration_scheme()

    fham = molecular_hamiltonian(water_data["ones"], water_data["twos"], constant_energy=water_data["constant_energy"])
    qham = tree.encode(fham)
    enumerated_qubit_hf_state = tree.hartree_fock_state(
        fermionic_hf_state=fermionic_hf_state,
    )
    assert np.isclose(PauliwordOp.from_dictionary(qham).expval(QuantumState([int(v) for v in enumerated_qubit_hf_state])), water_data["e_hf"])

@given(arrays(dtype=np.bool, shape=st.integers(1, 9)))
def test_naive_parity_hf_state(fermionic_hf_state):
    tree = ParityEncoding(len(fermionic_hf_state))
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    qubit_hf_state = tree.hartree_fock_state(
        fermionic_hf_state=fermionic_hf_state,
        mode_op_map=[*range(len(fermionic_hf_state))],
    )

    print(f"fermionic HF\t {fermionic_hf_state}")
    print(f"qubit HF\t {qubit_hf_state}")
    # The convention for Parity is that X is applied to indices
    # *higher* than the qubit being changed.
    # We have to change these around as we have an x-tail for lesser indices.
    expected_parity = np.cumsum(fermionic_hf_state[::-1]) % 2
    expected_parity = np.array(expected_parity, dtype=np.bool)[::-1]
    print(f"expected parity\t {expected_parity}")
    print(f"Result\t {np.all(qubit_hf_state == expected_parity)}")

    assert np.all(qubit_hf_state == expected_parity)

@given(arrays(dtype=np.bool, shape=st.integers(1, 9)))
def test_naive_jkmn_hf_state_runs(fermionic_hf_state):
    tree = JKMN(len(fermionic_hf_state))
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    qubit_hf_state = tree.hartree_fock_state(
        fermionic_hf_state=fermionic_hf_state,
        mode_op_map=[*range(len(fermionic_hf_state))],
    )

@pytest.mark.parametrize("encoding", [JW, PE, BK, JKMN])
def test_benchmark_encode_naive(benchmark,encoding, mol_data_sets):
    ones = mol_data_sets["ones"]
    twos = mol_data_sets["twos"]
    e_nuc = mol_data_sets["constant_energy"]
    fham = FermionHamiltonian(terms={"+-": ones, "++--": twos}, constant_energy=e_nuc)
    encoding = encoding(fham.n_modes)
    benchmark(lambda: encoding.encode(fham))

@pytest.mark.parametrize("encoding", [JW, PE, BK, JKMN])
def test_benchmark_encode_topphatt(benchmark,encoding, mol_data_sets):
    ones = mol_data_sets["ones"]
    twos = mol_data_sets["twos"]
    e_nuc = mol_data_sets["constant_energy"]
    fham = FermionHamiltonian(terms={"+-": ones, "++--": twos}, constant_energy=e_nuc)
    encoding:TernaryTree= encoding(fham.n_modes)
    benchmark(lambda: encoding.encode_topphatt(fham))


@pytest.mark.parametrize("encoding", [JW, PE, BK, JKMN])
def test_benchmark_encode_annealed(benchmark,encoding, h2_mol_data_sets):
    ones = h2_mol_data_sets["ones"]
    twos = h2_mol_data_sets["twos"]
    e_nuc = h2_mol_data_sets["constant_energy"]
    fham = FermionHamiltonian(terms={"+-": ones, "++--": twos}, constant_energy=e_nuc)
    encoding:TernaryTree = encoding(fham.n_modes)
    benchmark(lambda: encoding.encode_annealed(fham))


@pytest.mark.parametrize("encoding", [JW, PE, BK, JKMN])
@pytest.mark.parametrize("n_modes", [32,64,128])
def test_benchmark_hartree_fock_state(benchmark, encoding,n_modes):
    fermionic_hf_state = np.array([True]*n_modes, dtype=np.bool)
    tree:TernaryTree = encoding(len(fermionic_hf_state))
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    benchmark(lambda: tree.hartree_fock_state(fermionic_hf_state=fermionic_hf_state))


@st.composite
def tt_flatpack_strategy(draw, n_nodes_strategy):
    n_nodes = draw(n_nodes_strategy)
    if n_nodes == 0:
        return []
    nodes = draw(st.permutations([*range(n_nodes)]))
    flatpack = {n: {'x':None, 'y':None, 'z':None} for n in nodes}
    parents = [nodes[0]]
    for child in nodes[1:]:
        parent = draw(st.sampled_from(parents))
        assert isinstance(parent, int)
        parents.append(child)

        edges = [e for e, v in flatpack[parent].items() if v == None]
        flatpack[parent][draw(st.sampled_from(edges))] = child
        if len(edges) == 1:
            parents.remove(parent)
    # flatpack = [(k , tuple(val for val in v.values())) for k, v in flatpack.items()]
    to_add = [nodes[0]]
    output = []
    for node in to_add:
        output.append((node, tuple(v for v in flatpack[node].values())))
        for child in flatpack[node].values():
            if child is not None:
                to_add.append(child)
    return output


@given(flatpack=tt_flatpack_strategy(st.integers(1, 20)))
def test_validate_tt_flatpack_strategy(flatpack):
    used_qubit_indices = [flatpack[0][0]]
    for item in flatpack:
        assert item[0] in used_qubit_indices
        children = item[1]
        assert len(children) == 3
        for child in children:
            assert isinstance(child,  int | None)
            if isinstance(child, int):
                used_qubit_indices.append(child)


@pytest.mark.parametrize("encoding", [JW, PE, BK, JKMN])
@pytest.mark.parametrize("n_modes", [1, 5, 10])
def test_encoding_flatpack_validate(encoding, n_modes):
    tree = encoding(n_modes)
    flatpack = tree.flatpack()
    used_qubit_indices = [flatpack[0][0]]
    for item in flatpack:
        assert item[0] in used_qubit_indices
        children = item[1]
        assert len(children) == 3
        for child in children:
            assert isinstance(child,  int | None)
            if isinstance(child, int):
                used_qubit_indices.append(child)

@given(flatpack=tt_flatpack_strategy(st.integers(1, 20)))
def test_from_flatpack_roundtrip(flatpack):
    reconstructed = TernaryTree.from_flatpack(flatpack)
    assert sorted(reconstructed.flatpack()) == sorted(flatpack)


@given(flatpack=tt_flatpack_strategy(st.integers(1, 20)))
def test_from_flatpack_properties(flatpack):
    reconstructed = TernaryTree.from_flatpack(flatpack)

    # Check n_modes equals the number of nodes
    assert reconstructed.n_modes == len(flatpack)

    # Check n_qubits equals the number of nodes
    assert reconstructed.n_qubits == len(flatpack)

    # Check enumeration_scheme has correct length
    assert len(reconstructed.enumeration_scheme) == len(flatpack)

    # Check all qubit indices are unique and match flatpack ids
    qubit_indices = {mode_qubit[1] for mode_qubit in reconstructed.enumeration_scheme.values()}
    flatpack_ids = {item[0] for item in flatpack}
    assert qubit_indices == flatpack_ids


@given(flatpack=tt_flatpack_strategy(st.integers(1, 20)))
def test_symplectic_matrix_roundtrip(flatpack):
    original_tree = TernaryTree.from_flatpack(flatpack)
    print(f"Original flatpack {flatpack}")
    flatpack = original_tree.flatpack()
    reconstructed_tree = TernaryTree.from_flatpack(flatpack)
    print(f"Reconstructed flatpack {flatpack}")

    original_ipow, original_sym = original_tree._build_symplectic_matrix()
    reconstructed_ipow, reconstructed_sym = reconstructed_tree._build_symplectic_matrix()

    assert np.array_equal(original_ipow, reconstructed_ipow)
    assert np.array_equal(original_sym, reconstructed_sym)

@given(flatpack=tt_flatpack_strategy(st.integers(1, 20)))
def test_core_python_symplectics_from_flatpack_equal(flatpack):
    python_tree = TernaryTree.from_flatpack(flatpack)
    py_ipow, py_sym = python_tree._build_symplectic_matrix()
    rust_ipow, rust_sym = flatpack_symplectic_matrix(flatpack)

    assert np.array_equal(py_ipow, rust_ipow)
    assert np.array_equal(py_sym, rust_sym)



"""Visualisation of Majorana-string Encodings."""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import ListedColormap


def symplectic_matshow(symplectics, title: str | None = None):
    """Colorised Matplolibn.matshow of a symplectic array."""
    # Crear el mapa de colores cualitativo
    colors = ["linen", "tab:red", "tab:blue", "tab:purple"]

    left, right = np.hsplit(symplectics, 2)
    symplectics = (left + 2 * right) % 4

    # For cases where there is no Y or Z, the colormap must be scaled down.
    cmap = ListedColormap(colors[: np.max(symplectics) + 1])
    plt.matshow(symplectics, cmap=cmap, alpha=0.8)
    handles = [plt.Line2D([0], [0], color=color, lw=4) for color in colors]
    labels = ["I", "X", "Z", "Y"]
    plt.legend(handles, labels, loc="upper center", bbox_to_anchor=(0.5, 1.1), ncol=4)
    if title is not None:
        plt.title(title)
    plt.xlabel("Qubit")
    plt.tick_params(labeltop=False, labelbottom=True)
    plt.ylabel(r"Majorana $\gamma$")
    plt.yticks(range(symplectics.shape[0]))



"""Shared Fixtures for tests."""

import json
import numpy as np
from pytest import fixture
from ferrmion.encode import TernaryTree
from pathlib import Path
from openfermion import InteractionOperator, jordan_wigner, get_sparse_operator, QubitOperator
from scipy.sparse.linalg import eigsh
from ferrmion.utils import fermionic_to_sparse_majorana
from ferrmion.hamiltonians import FermionHamiltonian


@fixture(scope="module")
def water_data():
    folder = Path(__file__).parent
    with open(folder.joinpath("./data/h2o_sto-3g.json"), "rb") as file:
        data = json.load(file)
    data["ones"] = np.array(data["ones"])
    data["twos"] = np.array(data["twos"])
    return data

@fixture(scope="module", params=["h2_sto-3g","h2_6-31g"])
def h2_mol_data_sets(request):
    folder = Path(__file__).parent
    with open(folder.joinpath(f"./data/{request.param}.json"), "rb") as file:
        data = json.load(file)
    data["ones"] = np.array(data["ones"])
    data["twos"] = np.array(data["twos"])
    return data

@fixture(scope="module", params=["h2_sto-3g", "h2_6-31g", "water"])
def mol_data_sets(request):
    folder = Path(__file__).parent
    if request.param == "water":
        filename = "./data/h2o_sto-3g.json"
    else:
        filename = f"./data/{request.param}.json"
    with open(folder.joinpath(filename), "rb") as file:
        data = json.load(file)
    data["ones"] = np.array(data["ones"])
    data["twos"] = np.array(data["twos"])
    return data

@fixture(scope="module")
def water_tt(water_data) -> TernaryTree:
    return TernaryTree(14)

# @fixture(scope="module")
# def water_MaxNTO(water_data) -> MaxNTO:
#     return MaxNTO(*water_data)


@fixture(scope="module")
def water_eigenvalues(water_data) -> list[float]:
    return water_data["eigvals"]


@fixture(scope="module")
def water_fham(water_data) -> FermionHamiltonian:
    ones, twos = water_data
    return FermionHamiltonian(terms={"+-":water_data["ones"], "++--":water_data["twos"]})

@fixture(scope="module")
def water_sparse_majorana(water_data) -> dict:
    return fermionic_to_sparse_majorana(((water_data["ones"], "+-"), (water_data["twos"], "++--")))

def diagonalise_pauli_hamiltonian(qham, neigvals:int):
    ofop = QubitOperator()
    for k, v in qham.items():
        string = " ".join(
            [
                f"{char.upper()}{pos}" if char != "I" else ""
                for pos, char in enumerate(k)
            ]
        )
        ofop+= QubitOperator(term=string, coefficient=v)
    diag, _ = eigsh(get_sparse_operator(ofop), k=neigvals, which="SA")
    return diag

@fixture(scope="module")
def topphatt_weight_snapshot(water_data) -> dict:
    folder = Path(__file__).parent
    filename = "./data/topphatt_weight_snapshot.json"

    with open(folder.joinpath(filename), "rb") as file:
        data = json.load(file)

    return data



"""Tests for Interop Functions."""

from ferrmion.encode.ternary_tree import JordanWigner, BravyiKitaev
import pytest

try:
    import qiskit
    import qiskit_nature
    from qiskit_nature.second_q.mappers import JordanWignerMapper, BravyiKitaevMapper
    from qiskit_nature.second_q.operators import FermionicOp
    from qiskit.quantum_info import SparsePauliOp
    from ferrmion.interop.qiskit_mapper import QiskitAdapter
except ImportError:
    qiskit = None
    qiskit_nature = None


@pytest.mark.skipif(
    qiskit is None or qiskit_nature is None,
    reason="Extra depdency `ferrmion[qiskit]` not installed.",
)
def test_qiskit_adapter_jw():
    fop = FermionicOp(
        {"+_0 -_0": 1.0, "+_1 -_1": -1.0, "-_1 +_0": 2.0},
        num_spin_orbitals=2,
    )
    expected = SparsePauliOp(
        ["IZ", "ZI", "XY", "XX", "YY", "YX"],
        coeffs=[
            -0.5 + 0.0j,
            0.5 + 0.0j,
            0.0 + 0.5j,
            -0.5 + 0.0j,
            -0.5 + 0.0j,
            0.0 - 0.5j,
        ],
    )

    mapper = JordanWignerMapper()
    inbuilt = mapper.map(fop)
    assert inbuilt == expected
    assert expected.equiv(QiskitAdapter(JordanWigner(2)).map(fop))



"""Tests for TOPP-HATT Algorithm."""
from autoray import e

import json
import ferrmion as fr
from ferrmion.optimize.cost_functions import pauli_weight, coefficient_pauli_weight
from ferrmion.utils import fermionic_to_sparse_majorana
from ferrmion.encode.ternary_tree import (
    JordanWigner,
    BravyiKitaev,
    ParityEncoding,
    JKMN,
)
import numpy as np
import pytest
from ferrmion.core import topphatt_standard, encode
from ferrmion.optimize.huffman import huffman_ternary_tree
from ferrmion.optimize.hatt import hamiltonian_adaptive_ternary_tree, fast_hatt
from openfermion import QubitOperator, get_sparse_operator
from scipy.sparse.linalg import eigsh


@pytest.mark.parametrize("encoding", [JordanWigner, BravyiKitaev, ParityEncoding, JKMN])
def test_topphatt_preserves_topology(encoding, water_data):
    ones = water_data["ones"]
    twos = water_data["twos"]
    e_nuc = water_data["constant_energy"]
    fham = fr.molecular_hamiltonian(ones, twos, e_nuc)
    tree = encoding(14)
    _ = tree.encode_topphatt(fham)
    assert tree.root_node.child_strings == encoding(14).root_node.child_strings
    assert tree.root_node.branch_strings == encoding(14).root_node.branch_strings

def test_topphatt_huffman(water_data):
    ones = water_data["ones"]
    twos = water_data["twos"]
    e_nuc = water_data["constant_energy"]
    fham = fr.molecular_hamiltonian(ones, twos, e_nuc)
    test_tree = huffman_ternary_tree(water_data["ones"], water_data["twos"])
    initial_children = test_tree.root_node.child_strings
    initial_branches = test_tree.root_node.branch_strings
    _ = test_tree.encode_topphatt(fham)
    assert test_tree.root_node.child_strings == initial_children
    assert test_tree.root_node.branch_strings == initial_branches

def test_topphatt_hatt(water_data):
    ones = water_data["ones"]
    twos = water_data["twos"]
    e_nuc = water_data["constant_energy"]
    fham = fr.molecular_hamiltonian(ones, twos, e_nuc)
    test_tree = hamiltonian_adaptive_ternary_tree(fermionic_to_sparse_majorana(((water_data["ones"],"+-"), (water_data["twos"], "++--"))), n_modes=14)
    initial_children = test_tree.root_node.child_strings
    initial_branches = test_tree.root_node.branch_strings
    _ = test_tree.encode_topphatt(fham)
    assert test_tree.root_node.child_strings == initial_children
    assert test_tree.root_node.branch_strings == initial_branches

def test_topphatt_fasthatt(water_data):
    ones = water_data["ones"]
    twos = water_data["twos"]
    e_nuc = water_data["constant_energy"]
    fham = fr.molecular_hamiltonian(ones, twos, e_nuc)
    ones, twos = water_data["ones"], water_data["twos"]
    test_tree = fast_hatt(fermionic_to_sparse_majorana(((ones,"+-"), (twos, "++--"))), n_modes=14)
    initial_children = test_tree.root_node.child_strings
    initial_branches = test_tree.root_node.branch_strings
    _ =test_tree.encode_topphatt(fham)
    assert test_tree.root_node.child_strings == initial_children
    assert test_tree.root_node.branch_strings == initial_branches


@pytest.mark.parametrize("encoding", ["JW", "BK", "PE", "JKMN"])
@pytest.mark.parametrize("parallelize", [True, False])
def test_topphatt_standard_h2_eigvals_equal_expected(encoding, parallelize, h2_mol_data_sets):
    ones = h2_mol_data_sets["ones"]
    twos = h2_mol_data_sets["twos"]
    e_nuc = h2_mol_data_sets["constant_energy"]
    fham = fr.molecular_hamiltonian(ones, twos, e_nuc)
    tree = fr.TernaryTree(fham.n_modes)
    match encoding:
        case "JW":
            tree = tree.JW()
        case "BK":
            tree = tree.BK()
        case "PE":
            tree = tree.ParityEncoding()
        case "JKMN":
            tree = tree.JKMN()
    qham = tree.encode_topphatt(fham, parallelize)

    ofop = QubitOperator()
    for k, v in qham.items():
        string = " ".join(
            [
                f"{char.upper()}{pos}" if char != "I" else ""
                for pos, char in enumerate(k)
            ]
        )
        ofop+= QubitOperator(term=string, coefficient=v)

    print(expected:=h2_mol_data_sets["eigvals"])
    diag, _ = eigsh(get_sparse_operator(ofop), k=2*fham.n_modes, which="SA")
    print(diag)
    assert np.allclose(np.sort(diag), np.sort(h2_mol_data_sets["eigvals"]))

@pytest.mark.parametrize("encoding", ["JW", "BK", "PE", "JKMN","HATT", "Huffman"])
@pytest.mark.parametrize("parallelize", [True, False])
def test_topphatt_standard_h2o_weights_not_increased(encoding, parallelize, water_data, topphatt_weight_snapshot):
    ones = water_data["ones"]
    twos = water_data["twos"]
    e_nuc = water_data["constant_energy"]
    fham = fr.molecular_hamiltonian(ones, twos, e_nuc)
    tree = fr.TernaryTree(fham.n_modes)
    match encoding:
        case "JW":
            tree = tree.JW()
        case "BK":
            tree = tree.BK()
        case "PE":
            tree = tree.ParityEncoding()
        case "JKMN":
            tree = tree.JKMN()
        case "HATT":
            tree = hamiltonian_adaptive_ternary_tree(
                fr.core.fermionic_to_sparse_majorana(*fham.signatures_and_coefficients, fham.constant_energy),
                fham.n_modes)
        case "Huffman":
            tree = huffman_ternary_tree(ones, twos)
    qham_naive = tree.encode(fham)
    qham = tree.encode_topphatt(fham, parallelize)

    assert np.isclose(float(pauli_weight(qham)[0]/pauli_weight(qham_naive)[0]),topphatt_weight_snapshot[encoding]["pauli_weight"], atol=0.01)
    assert np.isclose(float(coefficient_pauli_weight(qham)[0]/coefficient_pauli_weight(qham_naive)[0]), topphatt_weight_snapshot[encoding]["coefficient_pauli_weight"], atol=0.01)



"""Base FermionQubitEncoding class."""

import logging
from abc import ABC, abstractmethod
from typing import Callable

import numpy as np
from numpy.typing import ArrayLike, NDArray

from ferrmion.core import (
    anneal_enumerations,
    encode,
    encode_fermion_product,
    symplectic_product_map,
)
from ferrmion.hamiltonians import FermionHamiltonian, QubitHamiltonian
from ferrmion.utils import (
    pauli_to_symplectic,
    symplectic_to_pauli,
)

logger = logging.getLogger(__name__)


class FermionQubitEncoding(ABC):
    """Fermion Encodings for the Electronic Structure Hamiltonian in symplectic form.

    Attributes:
        one_e_coeffs (NDArray): One electron coefficients.
        two_e_coeffs (NDArray): Two electron coefficients.
        modes (set[int]): A set of modes.
        n_qubits (int): The number of qubits.

    Methods:
        default_mode_op_map: Get the default mode operator map.
        _build_symplectic_matrix: Build a symplectic matrix representing terms for each operator in the Hamiltonian.
        hartree_fock_state: Find the Hartree-Fock state of a majorana string encoding.
        _symplectic_to_pauli: Convert a symplectic matrix to a Pauli string.
        _pauli_to_symplectic: Convert a Pauli string to a symplectic matrix.
        fill_template: Fill a template with Hamiltonian coefficients.
        to_symplectic_hamiltonian: Output the hamiltonian in symplectic form.
        to_qubit_hamiltonian: Create qubit representation Hamiltonian.

    NOTE: A 'Y' pauli operator is mapped to -iXY so a (0+n)**3 term is needed.

    Example:
        >>> from ferrmion.encode.base import FermionQubitEncoding
        >>> class DummyEncoding(FermionQubitEncoding):
        ...     def _build_symplectic_matrix(self):
        ...         import numpy as np
        ...         return np.zeros(1), np.zeros((1, 2))
        >>> enc = DummyEncoding(2, 2)
        >>> enc.n_modes
        2
    """

    def __init__(
        self,
        n_modes: int,
        n_qubits: int,
    ):
        """Initialise encoding.

        Args:
            n_modes (int): Number of Fermion modes to encode.
            n_qubits (int): Number of Qubits used to encode.
            vacuum_state (NDArray | None): The vacuum state of the encoding.
        """
        self.n_modes = n_modes
        self.n_qubits = n_qubits
        self.default_mode_op_map = np.array([*range(self.n_modes)], dtype=np.uint)

    def __eq__(self, other: object) -> bool:
        """Checks if two encodings are exactly equivalent."""
        if isinstance(other, FermionQubitEncoding):
            if self.n_modes != other.n_modes:
                return False

            if self.n_qubits != other.n_qubits:
                return False

            left = self._build_symplectic_matrix()
            right = other._build_symplectic_matrix()

            if not np.all(left[0] == right[0]) or not np.all(left[1] == right[1]):
                return False

            return True
        else:
            return False

    @property
    def default_mode_op_map(self) -> NDArray[np.uint]:
        """Create a default mode operator map for the tree."""
        return self._default_mode_op_map

    @default_mode_op_map.setter
    def default_mode_op_map(self, permutation: list[int]):
        """Set the default mode operator map.

        Args:
            permutation (list[int]): A list containing a permutation of mode indices.
        """
        logger.debug("Setting default mode operator map.")
        error_string = ""
        if set(permutation) != {*range(self.n_modes)}:
            error_string += "Default Mode op map does not cover all modes.\n"

        if error_string != "":
            logger.error(error_string)
            logger.error(permutation)
            raise ValueError(error_string)

        self._default_mode_op_map = np.array(permutation, dtype=np.uint)

    @property
    def vacuum_state(self):
        """Return the vacuum state."""
        return self._vacuum_state

    @vacuum_state.setter
    def vacuum_state(self, state: NDArray):
        """Validate and set the vacuum state.

        Args:
            state (NDArray): The vacuum state.
        """
        logger.debug("Setting vacuum state as %s", state)
        error_string = []
        state = np.array(state, dtype=np.float64)

        if len(state) != self.n_qubits:
            error_string.append("vacuum state must be length " + str(self.n_qubits))
        if state.ndim != 1:
            error_string.append("vacuum state must be vector (dimension==1)")

        if error_string != []:
            logger.error("\n".join(error_string))
            raise ValueError("\n".join(error_string))
        else:
            self._vacuum_state = state

    @abstractmethod
    def _build_symplectic_matrix(
        self,
    ) -> tuple[NDArray[np.uint8], NDArray[bool]]:
        """Build a symplectic matrix representing terms for each operator in the Hamitonian."""
        pass

    def encode_annealed(
        self,
        fham: FermionHamiltonian,
        temperature: int | None = None,
        initial_guess: list[int] | None = None,
        coefficient_weighted: bool = True,
    ):
        """Encode a Hamiltonian, using simulated annealing optimisation."""
        sigs, coeffs = fham.signatures_and_coefficients
        ipow, sym = self._build_symplectic_matrix()

        if temperature is None:
            temperature = fham.n_modes

        if isinstance(initial_guess, list):
            initial_guess: NDArray[np.uint] = np.array(initial_guess, dtype=np.uint)
        else:
            initial_guess: NDArray[np.uint] = np.array(
                [*range(self.n_modes)], dtype=np.uint
            )

        ipow, sym = anneal_enumerations(
            ipowers=ipow,
            symplectics=sym,
            signatures=sigs,
            coeffs=coeffs,
            temperature=temperature,
            initial_guess=initial_guess,
            coefficient_weighted=coefficient_weighted,
        )

        self._build_symplectic_matrix: Callable = lambda: (ipow, sym)
        self.default_mode_op_map = [*range(self.n_modes)]

        return encode(
            ipowers=ipow,
            symplectics=sym,
            signatures=sigs,
            coeffs=coeffs,
            constant_energy=fham.constant_energy,
        )

    def to_json(self) -> dict:
        """Store the encoding as a JSON-readable dict."""
        ipow, sym = self._build_symplectic_matrix()
        dict_output = {}
        dict_output["ipowers"] = ipow.tolist()
        dict_output["symplectics"] = sym.tolist()
        return dict_output

    def encode(self, fham: FermionHamiltonian) -> QubitHamiltonian:
        """Encode a Hamiltonian."""
        logger.debug("Encoding fermionic Hamiltonian.")
        ipowers, symplectic = self._build_symplectic_matrix()
        signatures, coeffs = fham.signatures_and_coefficients
        return encode(
            ipowers=ipowers,
            symplectics=symplectic,
            signatures=signatures,
            coeffs=coeffs,
            constant_energy=fham.constant_energy,
        )

    @staticmethod
    def _symplectic_to_pauli(
        symplectic: NDArray,
        ipower: int = 0,
    ) -> tuple[str, int]:
        """Convert a symplectic matrix to a Pauli string.

        Args:
            ipower (NDArray[np.uint]): power of i coefficient
            symplectic (NDArray): A symplectic vector.
        """
        return symplectic_to_pauli(symplectic, ipower)

    @staticmethod
    def _pauli_to_symplectic(
        pauli: str,
        ipower: int = 0,
    ) -> tuple[NDArray[bool], int]:
        """Convert a Pauli string to a symplectic matrix.

        Args:
            ipower (NDArray[np.uint]): power of i coefficient
            pauli (str): A Pauli-string.
        """
        return pauli_to_symplectic(pauli, ipower)

    @property
    def symplectic_product_map(self):
        """Calculate the product of symplectic terms and cache them."""
        logger.debug("Building symplectic product map")
        ipowers, symplectics = self._build_symplectic_matrix()
        return symplectic_product_map(ipowers, symplectics)

    def number_operator(
        self,
        mode: int,
        coeff: float | np.complex64 = np.complex64(1.0, 0.0),
    ) -> QubitHamiltonian:
        """Return the number operator of a mode for this encoding.

        Args:
            mode (int): The mode index to obtain a number operator for.
            coeff (float): Optional complex coeffient

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(4)
            >>> n_zero = tree.number_operator(0)
        """
        return self._encode_fermion_product(
            signature="+-",
            mode_indices=[mode, mode],
            coeff=coeff,
        )

    def edge_operator(
        self,
        edge_indices: tuple[int, int],
        coeff: float | np.complex64 = np.complex64(1.0, 0.0),
    ) -> QubitHamiltonian:
        """Return the edge operator of a pair of modes for this encoding.

        Args:
            edge_indices (tuple[int, int]): The mode index to obtain a number operator for.
            coeff (float): Optional complex coeffient

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(4)
            >>> tree.edge_operator((0, 1))
        """
        return self._encode_fermion_product(
            signature="+-",
            mode_indices=list(edge_indices),
            coeff=coeff,
        )

    def interaction_operator(
        self,
        mode_indices: tuple[int, int, int, int],
        coeff: float | np.complex64 = np.complex64(1.0, 0.0),
        physicist_notation=True,
    ) -> QubitHamiltonian:
        """Return an interaction operator of four modes for this encoding.

        Args:
            mode_indices (tuple[int, int, int, int]): The mode index to obtain an interaction operator for.
            coeff (float): Optional complex coeffient
            physicist_notation (bool): Use Physicist's notation, Chemist's if False.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(4)
            >>> tree.interaction_operator((0, 1, 2, 3))
        """
        if physicist_notation:
            return self._encode_fermion_product("++--", list(mode_indices), coeff=coeff)
        else:
            return self._encode_fermion_product("+-+-", list(mode_indices), coeff=coeff)

    def _encode_fermion_product(
        self,
        signature: str,
        mode_indices: list[int],
        coeff: np.complex64 | float,
    ) -> QubitHamiltonian:
        """Returns the sparse pauli form of a double fermionic operator.

        Args:
            signature (str): The fermionic operator signature, composed of "+" and "-".
            mode_indices (list[int]): The mode indices to obtain a number operator for.
            coeff (float): The operator coefficient.

        Returns:
            list[tuple[str, NDArray, np.complexfloating]]: A list of tuples each containing a Pauli string, its qubit indices and a complex coefficient.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(4)
            >>> _encode_fermion_product(tree, "++--", (0, 1, 2, 3), 1.0)
        """
        if not all([ind in range(self.n_modes) for ind in mode_indices]):
            raise ValueError("Indices invalid.")
        if len(signature) != len(mode_indices):
            raise ValueError("Signature and indices must be same length.")
        if isinstance(coeff, float | int):
            coeff = np.complex64(coeff, 0.0)
        return encode_fermion_product(
            *self._build_symplectic_matrix(), signature, mode_indices, coeff
        )


class MajoranaStringEncoding(FermionQubitEncoding):
    """Majorana string encoding defined by input."""

    def __init__(self, ipowers: ArrayLike, symplectics: ArrayLike):
        """Initialize MajoranaStringEncoding.

        Args:
            ipowers (ArrayLike): Array of integers representing powers of the imaginary unit (i).
            symplectics (ArrayLike): Array of booleans representing Pauli strings in XZ format.
        """
        self._ipowers = np.array(ipowers, dtype=np.uint8)
        self._ipowers %= 4

        self._symplectics = np.array(symplectics, dtype=np.bool)

        if len(self._ipowers) != len(self._symplectics):
            raise ValueError("ipowers and symplectics must be same length.")
        self.n_modes = self._symplectics.shape[0] // 2
        self.n_qubits = self._symplectics.shape[1] // 2
        super().__init__(self.n_modes, self.n_qubits)

    def _build_symplectic_matrix(
        self,
    ) -> tuple[NDArray[np.uint8], NDArray[np.bool]]:
        """Build the symplectic matrix for the Majorana string encoding.

        Returns:
            Tuple[ArrayLike, ArrayLike]: The symplectic matrix in XZ format.
        """
        return self._ipowers, self._symplectics



"""Bonsai Algorithm."""

import logging

import numpy as np
import rustworkx as rx

from ferrmion import TernaryTree
from ferrmion.encode.ternary_tree_node import TTNode

logger = logging.getLogger(__name__)


def bonsai_algorithm(
    graph: rx.PyGraph, homogenous: bool = True, max_nodes: None | int = None
) -> TernaryTree:
    """Create a TernayTree encoding using the Bonsai Algorithm.

    Args:
        graph (rx.PyGraph): A RustworkX graph of device qubit-connectivity.
        homogenous (bool): "homogenous" labelling if true, else "heterogenous"
        max_nodes (int): Maximum number of nodes to include in output tree.

    Returns:
        TernaryTree: A ternary tree encoding.
    """
    logger.debug("Starting Bonsai Algorithm.")
    if homogenous:
        chars = ["x", "y", "z"]
    else:
        chars = ["z", "x", "y"]

    distances = rx.distance_matrix(graph)
    root_index = int(np.argmin(np.max(distances, axis=1)))
    node_queue = [root_index]
    used_indices = {root_index}
    nodes = [TTNode(parent=None) for _ in range(graph.num_nodes())]
    nodes[root_index].root_path = ""
    nodes[root_index].qubit_label = root_index

    while len(node_queue) > 0:
        logger.debug(f"{node_queue=}")
        logger.debug(f"{used_indices=}")
        node: int = node_queue.pop(0)
        parent = nodes[node]
        logger.debug(node)

        neighbors = sorted(list(set(graph.neighbors(node)).difference(used_indices)))
        logger.debug(f"{neighbors=}")

        n_neighbors = len(neighbors)

        for neighbor, char in zip(neighbors[:3], chars[:n_neighbors]):
            node_queue.append(neighbor)
            used_indices.add(neighbor)

            parent.add_child(
                char,
                child_node=nodes[neighbor],
                root_path=f"{parent.root_path}{char}",
                qubit_label=neighbor,
            )
            nodes[neighbor].parent = parent

            if max_nodes is not None and len(used_indices) == max_nodes:
                break

        if max_nodes is not None and len(used_indices) == max_nodes:
            break
        logger.debug(node_queue)
        logger.debug("")

    if max_nodes is not None and len(used_indices) == max_nodes:
        logger.debug(f"Found {max_nodes} nodes.")
    elif len(used_indices) == graph.num_nodes():
        logger.debug("Found spanning tree")
    else:
        logger.debug("Tree does not span the graph.")
        unused_indices = set(range(graph.num_nodes())).difference(used_indices)
        for unused in unused_indices:
            closest = np.argsort(distances[unused])
            for used in closest:
                used_node: TTNode = nodes[used]
                for child_branch in chars:
                    if getattr(used_node, child_branch) is None:
                        used_node.add_child(
                            child_branch,
                            child_node=nodes[unused],
                            root_path=f"{used_node.root_path}{child_branch}",
                            qubit_label=unused,
                        )
                        unused_indices.remove(used)
                        break

        if len(unused_indices) > 0:
            logger.debug("Error, not all qubits assigned to nodes.")
        else:
            logger.debug("All graph nodes assigned to tree.")

    logger.debug("Creating encoding.")

    n_nodes = graph.num_nodes() if max_nodes is None else max_nodes
    tree = TernaryTree(
        n_modes=n_nodes, n_qubits=graph.num_nodes(), root_node=nodes[root_index]
    )
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    return tree



"""Tests for base fermion to qubit encoding class"""
from typing import Callable
from IPython.utils.data import T
from autoray.lazy import exp
from jupyter_lsp.specs import r

import numpy as np
import pytest
from ferrmion.encode import TernaryTree, MaxNTO
from ferrmion.encode.ternary_tree import JordanWigner, BravyiKitaev, JKMN, ParityEncoding
from ferrmion.encode.base import FermionQubitEncoding, MajoranaStringEncoding
from hypothesis import strategies as st, given

np.random.seed(1710)


@pytest.fixture
def four_mode_tt():
    return TernaryTree(n_modes=4)


@pytest.fixture
def sixteen_mode_tt():
    return TernaryTree(n_modes=16)


def test_default_vacuum_state(four_mode_tt):
    assert np.all(four_mode_tt.vacuum_state == np.array([0] * 4))


def test_valid_vacuum_state(four_mode_tt):
    with pytest.raises(ValueError) as excinfo:
        four_mode_tt.vacuum_state = [0] * 3
    assert "4" in str(excinfo.value)
    assert "length" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        four_mode_tt.vacuum_state = [0] * 5
    assert "4" in str(excinfo.value)
    assert "length" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        four_mode_tt.vacuum_state = np.array([[0], [0]])
    assert "dimension" in str(excinfo.value)


def test_hartree_fock_state(sixteen_mode_tt):
    jw = sixteen_mode_tt.JW()
    hartree_fock_state = jw.hartree_fock_state
    nq = jw.n_qubits // 2
    print(hartree_fock_state(np.array([True] * nq + [False] * nq, dtype=bool)))
    assert np.all(
        hartree_fock_state(np.array([True] * nq + [False] * nq, dtype=bool))
        == np.array([[True] * nq + [False] * nq], dtype=bool)
    )
    assert np.all(
        hartree_fock_state(
            np.array([True] * (nq + 1) + [False] * (nq - 1), dtype=bool)
        )
        == np.array([[True] * (nq + 1) + [False] * (nq - 1)], dtype=bool)
    )


@pytest.mark.parametrize("encoding", [JordanWigner, ParityEncoding, BravyiKitaev, JKMN])
@given(n_modes=st.integers(min_value=1, max_value=10), operator_mode=st.integers(min_value=1, max_value=10))
def test_number_operator_equals_edge_operator(encoding: Callable[[int], TernaryTree], n_modes:int, operator_mode:int):
    tree = encoding(n_modes)
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    if operator_mode < n_modes:
        # numpy doesn't like comparing empty arrays
        assert tree.edge_operator((operator_mode, operator_mode)) == tree.number_operator(operator_mode)
    else:
        with pytest.raises(ValueError) as excinfo:
            tree.number_operator(tree.n_modes + 1)
        assert "Indices invalid" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        tree.number_operator(-1)
        assert "Indices invalid" in str(excinfo.value)


def test_edge_operator(four_mode_tt: TernaryTree):
    tree: TernaryTree = four_mode_tt.JKMN()
    tree.enumeration_scheme = tree.default_enumeration_scheme()

    output = tree.edge_operator((0, 3))
    expected = {"YZIX": -0 - 0.25j,"YZIY":0.25 - 0j, "XIZX":0.25 + 0j, "XIZY":0 + 0.25j,
    }

    assert output == expected

    scaled_output = tree.edge_operator((0,3), 0.5)
    expected = {k:0.5*v for k,v in output.items()}

    assert scaled_output == expected

@given(scaler=st.complex_numbers(min_magnitude=1e-2, max_magnitude=1e5))
def test_encode_fermion_product_coefficient_scaling_correct(scaler:np.complex64):
    four_mode_tt = TernaryTree(4)

    jw_expected = {"IIII":0.5, "ZIII":-0.5}
    jw_num_zero = four_mode_tt.JW()._encode_fermion_product("+-", [0,0], 1.)
    jw_num_zero_scaled = four_mode_tt.JW()._encode_fermion_product("+-", [0,0], scaler)
    assert jw_expected==jw_num_zero
    scaler_expected = {k:scaler*v for k,v in jw_expected.items()}
    assert set(jw_num_zero_scaled.keys()) == set(scaler_expected.keys())
    assert all([np.isclose(scaler_expected[k], jw_num_zero_scaled[k]) for k in jw_num_zero_scaled.keys()])

    bk_expected = {"IIII": 0.5, "ZZIZ":-0.5}
    bk_num_zero = four_mode_tt.BK()._encode_fermion_product("+-", [0,0],1.)
    bk_num_zero_scaled = four_mode_tt.BK()._encode_fermion_product("+-", [0,0],scaler)
    assert bk_expected==bk_num_zero
    scaler_expected = {k:scaler*v for k,v in bk_expected.items()}
    assert set(bk_num_zero_scaled.keys()) == set(scaler_expected.keys())
    assert all([np.isclose(scaler_expected[k], bk_num_zero_scaled[k]) for k in bk_num_zero_scaled.keys()])

    maxnto_expected = {"IIII":0.5, "IZZZ":0.5}
    maxnto_num_zero = MaxNTO(4)._encode_fermion_product( "+-", [0,0],1.)
    maxnto_num_zero_scaled = MaxNTO(4)._encode_fermion_product( "+-", [0,0],scaler)
    assert maxnto_expected == maxnto_num_zero
    scaler_expected = {k:scaler*v for k,v in maxnto_expected.items()}
    assert set(maxnto_num_zero_scaled.keys()) == set(scaler_expected.keys())
    assert all([np.isclose(scaler_expected[k], maxnto_num_zero_scaled[k]) for k in maxnto_num_zero_scaled.keys()])



"""Tests for TTNode class."""

from ferrmion.encode.ternary_tree_node import (
    TTNode,
    z_ancestor,
    z_descendant,
    node_sorter,
)
from ferrmion.encode.ternary_tree import JKMN
import numpy as np


def test_ttnode_update_root_path():
    root = TTNode()
    assert root.root_path == ""
    xchild = root.add_child("x")
    assert xchild.root_path == "x"
    ychild = root.add_child("y")
    assert ychild.root_path == "y"
    yxchild = ychild.add_child("x")
    assert yxchild.root_path == "yx"

    root.update_root_path("xyz")
    assert root.root_path == "xyz"
    assert xchild.root_path == "xyzx"
    assert ychild.root_path == "xyzy"
    assert yxchild.root_path == "xyzyx"


def test_ttnode_z_relatives():
    root = TTNode()
    child = root.add_child("z")
    assert child.root_path == "z"
    grandchild = child.add_child("z")
    assert grandchild.root_path == "zz"

    # descendant
    assert z_descendant(root) is grandchild
    assert z_descendant(child) is grandchild

    # ancestor
    assert z_ancestor(child) is root
    assert z_ancestor(grandchild) is root


def test_ttnode_add_child():
    root = TTNode()
    child = root.add_child("x")
    child = child.add_child("y")
    child = child.add_child("z")
    assert root.parent is None
    assert child == root.x.y.z
    assert child.parent == root.x.y
    assert child.parent.parent == root.x
    assert child.parent.parent.parent == root
    assert root.root_path == ""
    assert root.x.root_path == "x"

    new_root = TTNode()
    moved_child = new_root.add_child(which_child="x", child_node=root.x.y)
    assert root.x.y is None
    assert new_root.x is moved_child
    assert moved_child.root_path == "x"
    assert isinstance(moved_child.z, TTNode)
    assert moved_child.z.root_path == "xz"


def test_ttnode_as_dict():
    root = TTNode()
    child = root.add_child("x")
    child = child.add_child("y")
    child = child.add_child("z")
    assert root.as_dict() == {
        "x": {
            "x": None,
            "y": {"x": None, "y": None, "z": {"x": None, "y": None, "z": None}},
            "z": None,
        },
        "y": None,
        "z": None,
    }


def test_ttnode_child_strings():
    root = TTNode()
    child = root.add_child("x")
    child = child.add_child("y")
    child = child.add_child("z")
    assert root.child_strings == [
        "",
        "x",
        "xy",
        "xyz",
    ]


def test_ttnode_child_qubit_labels():
    root = TTNode()
    child = root.add_child("x")
    child = child.add_child("y")
    child = child.add_child("z")
    assert root.child_qubit_labels == {"": None, "x": None, "xy": None, "xyz": None}

    root.qubit_label = 1
    root.x.qubit_label = 0
    root.x.y.qubit_label = 3
    root.x.y.z.qubit_label = 2
    assert root.child_qubit_labels == {"": 1, "x": 0, "xy": 3, "xyz": 2}


def test_ttnode_branch_strings():
    root = TTNode()
    child = root.add_child("x")
    child = child.add_child("y")
    child = child.add_child("z")
    assert root.branch_strings == {
        "xx",
        "xyx",
        "xyy",
        "xyzx",
        "xyzy",
        "xyzz",
        "xz",
        "y",
        "z",
    }


def test_ttnode_node_sorter():
    assert node_sorter("z") == 3
    assert node_sorter("xx") == 11
    assert node_sorter("xyz") == 123


def test_ttnode_branch_majorana_map():
    root = TTNode()
    child = TTNode()
    assert root.leaf_majorana_indices == {"x": None, "y": None, "z": None}
    assert child.leaf_majorana_indices == {"x": None, "y": None, "z": None}
    child.leaf_majorana_indices = {"x": 0, "y": 1, "z": 2}

    root.add_child(which_child="x", child_node=child, root_path=None, qubit_label=None)
    assert root.branch_majorana_map == {"y": None, "z": None, "xx": 0, "xy": 1, "xz": 2}


def test_ttnode_to_rustworkx():
    graph = JKMN(6).root_node.to_rustworkx()
    assert graph.nodes() == ["", "x", "y", "z", "xx", "xy"]
    assert np.all([*graph.edge_list()] == [(0, 1), (0, 2), (0, 3), (1, 4), (1, 5)])



import numpy as np
from ferrmion.core import symplectic_product, topphatt, topphatt_standard, encode, encode_standard, standard_symplectic_matrix
import pytest

def test_symplectic_product():
    xyz = np.array([1, 1, 0, 0, 1, 1], dtype=bool)
    xxx = np.array([1, 1, 1, 0, 0, 0], dtype=bool)
    zzz = np.array([0, 0, 0, 1, 1, 1], dtype=bool)
    yyy = np.array([1, 1, 1, 1, 1, 1], dtype=bool)
    yzx = np.array([1, 0, 1, 1, 1, 0], dtype=bool)
    assert symplectic_product(xxx, zzz)[0] == 0
    assert np.all(symplectic_product(xxx, zzz)[1] == np.array([1, 1, 1, 1, 1, 1]))
    assert symplectic_product(zzz, xxx)[0] == 2
    assert np.all(symplectic_product(zzz, xxx)[1] == np.array([1, 1, 1, 1, 1, 1]))

    assert symplectic_product(xxx, yyy)[0] == 0
    assert np.all(symplectic_product(xxx, yyy)[1] == np.array([0, 0, 0, 1, 1, 1]))
    assert symplectic_product(yyy, xxx)[0] == 2
    assert np.all(symplectic_product(yyy, xxx)[1] == np.array([0, 0, 0, 1, 1, 1]))

    assert symplectic_product(zzz, yyy)[0] == 2
    assert np.all(symplectic_product(zzz, yyy)[1] == np.array([1, 1, 1, 0, 0, 0]))
    assert symplectic_product(yyy, zzz)[0] == 0
    assert np.all(symplectic_product(yyy, zzz)[1] == np.array([1, 1, 1, 0, 0, 0]))

    assert symplectic_product(xxx, xyz)[0] == 0
    assert np.all(symplectic_product(xxx, xyz)[1] == np.array([0, 0, 1, 0, 1, 1]))
    assert symplectic_product(xyz, xxx)[0] == 0
    assert np.all(symplectic_product(xyz, xxx)[1] == np.array([0, 0, 1, 0, 1, 1]))

    assert symplectic_product(yzx, xyz)[0] == 0
    assert np.all(symplectic_product(yzx, xyz)[1] == np.array([0, 1, 1, 1, 0, 1]))
    assert symplectic_product(xyz, yzx)[0] == 2
    assert np.all(symplectic_product(xyz, yzx)[1] == np.array([0, 1, 1, 1, 0, 1]))

def test_core_topphatt_flatpack_runs(water_data):
    flatpack = [(i, (None, None, i+1)) for i in range(13)] + [(13, (None, None, None))]
    topphatt(flatpack,14, signatures=["+-", "++--"], coeffs=[water_data["ones"], water_data["twos"]], parallelize=True)

@pytest.mark.parametrize("encoding", ["JW", "BK", "PE", "JKMN"])
def test_core_standard(encoding, water_eigenvalues, water_data):
    ones = water_data["ones"]
    twos = water_data["twos"]
    one_step = encode_standard(encoding, 14,14, ["+-","++--"], [ones, twos], 0.)

    ipow, sym = standard_symplectic_matrix(encoding, ones.shape[0])
    two_step = encode(ipow, sym,["+-","++--"], [ones, twos], 0.)

    assert len(one_step) == len(two_step)
    for k,v in one_step.items():
        assert two_step[k] == v



"""Build the weird k-NTO encodings."""

import logging

import numpy as np
from numpy.typing import NDArray

from ferrmion.utils import xy_swap

from .base import FermionQubitEncoding

logger = logging.getLogger(__name__)


class MaxNTO(FermionQubitEncoding):
    """k-NTO encoding for fermionic operators.

    Attributes:
        n_modes (int): The number of modes.

    Methods:
        _build_symplectic_matrix(): Build the symplectic matrix for the k-NTO encoding.
        _valid_qubit_number(): Check if the number of qubits is valid for the k-NTO encoding.
    """

    def __init__(self, n_modes):
        """Initialise a k-NTO encoding.

        Args:
            n_modes (int): The number of fermionic modes
        """
        self.n_modes = n_modes
        super().__init__(n_modes=n_modes, n_qubits=n_modes)

    def _build_symplectic_matrix(self) -> tuple[NDArray[np.number], NDArray[bool]]:
        """Build the symplectic matrix for the k-NTO encoding.

        Returns:
            NDArray: The symplectic matrix.

        Example:
            >>> from ferrmion.encode.maxnto import MaxNTO
            >>> MaxNTO = MaxNTO(5)
            >>> y_count, sympl = MaxNTO._build_symplectic_matrix()
        """
        return maxnto_symplectic_matrix(self.n_modes)

    def _valid_qubit_number(self) -> int:
        """Check if the number of qubits is valid for the k-NTO encoding.

        Returns:
            int: The number of qubits.
        """
        return self.n_modes


def maxnto_symplectic_matrix(n_modes) -> tuple[NDArray[np.number], NDArray[bool]]:
    """Build a symplectic matrix of majorana operators for the k-NTO encoding.

    Args:
        n_modes (int): The number of modes.

    Returns:
        tuple[NDArray, NDArray]: The y_count of each vector and the symplectic matrix.

    Example:
        >>> from ferrmion.encode.maxnto import maxnto_symplectic_matrix
        >>> y_count, sympl = maxnto_symplectic_matrix(5)
        >>> sympl.shape
    """
    logger.debug(f"Building k-NTO symplectic matrix for {n_modes=}")
    k = n_modes - 1
    if k % 2 != 1:
        raise ValueError("Only works for Odd k")

    # Choice of x_block and z_block is arbitary but at least for TNs
    # having the simple block on the z_block was better.
    x_block = np.zeros((n_modes, n_modes))
    x_block += np.triu(np.ones(n_modes), k=1)
    x_block += np.tril(np.ones(n_modes), k=-1)

    z_block = np.tril(np.ones(n_modes), k=-1)
    for i in range(0, z_block.shape[0], 2):
        z_block[i, i] = True

    # Y = iXZ
    x_block = np.array(x_block, dtype=bool)
    z_block = np.array(z_block, dtype=bool)
    z_block[1::2] = z_block.T[1::2, :]
    odd_majoranas = np.empty((n_modes, n_modes), dtype=np.uint8)
    even_majoranas = np.empty(x_block.shape, dtype=np.uint8)
    odd_majoranas = np.hstack((x_block, z_block), dtype=np.uint8)
    even_majoranas = np.hstack((x_block, z_block), dtype=np.uint8)
    even_majoranas = xy_swap(even_majoranas)

    output = np.empty((2 * n_modes, 2 * n_modes), dtype=bool)
    output[::2, :] = odd_majoranas
    output[1::2, :] = even_majoranas
    output = np.array(output, dtype=bool)
    y_count = np.zeros(output.shape[0], dtype=np.uint8)
    y_count = (
        np.sum(
            np.bitwise_and(output[:, :n_modes], output[:, n_modes:]),
            axis=1,
            dtype=np.uint8,
        )
        % 4
    )
    return y_count, output



"""Classes which represent physical devices or objects."""

import logging
from abc import ABC

logger = logging.getLogger(__name__)


class Qubit(ABC):
    """A qubit object which represents a physical qubit.

    Args:
        label (int): The qubit label.
        gate_error (float): The gate error rate.
        t1 (float): The T1 time.
        t2 (float): The T2 time.
    """

    label: int
    gate_error: float
    t1: float
    t2: float


class Toplogy:
    """A topology object which represents a physical device.

    Attributes:
        qubits (set[Qubit]): A set of qubits.
        connections (dict): A dictionary of qubit connections.

    Methods:
        add_connection(control, target, error): Add a connection between two qubits.
    """

    def __init__(self, qubits: set[Qubit]):
        """Initialize the topology object.

        Args:
            qubits (set[Qubit]): A set of qubits.
        """
        self.qubits = qubits
        self.connections = {q.root_path: {} for q in qubits}

    def add_connection(self, control: Qubit, target: Qubit, error: float):
        """Add a connection between two qubits.

        Args:
            control (Qubit): The control qubit.
            target (Qubit): The target qubit.
            error (float): The error rate of the connection.
        """
        # check if the control is in the set, then check if a value is set for the target error
        self.connections[control.root_path][target.root_path] = error



"""Evolutionary Algorihm Based Optimizations."""

import random
from functools import partial

import numpy as np
from deap import algorithms, base, creator, tools
from numpy.typing import NDArray


def lambda_plus_mu(
    n_modes: int,
    evaluate: partial[list[float]],
    pop_size: int,
    ngen: int,
) -> NDArray:
    """Reorder modes using lambda+mu evolutionary algorithm.

    The values of lambda and mu are both set to half of the population size.
    With some tinkering this seems to work well.

    Args:
        n_modes (int): The number of modes in the system.
        evaluate (partial[list[float]]): A partial function which acts as the cost function.
            This should take a list of integers (mode labels) as input
            and output a list of floats.
        pop_size (int): The size of the initial population.
        ngen (int): The number of generations to evolve.

    Returns:
        NDArray: The best mode ordering found.

    Example:
        >>> import numpy as np
        >>> from functools import partial
        >>> from ferrmion.optimize.enumeration.evolutionary import lambda_plus_mu
        >>> def dummy_eval(x): return [sum(x)]
        >>> evaluate = partial(dummy_eval)
        >>> lambda_plus_mu(3, evaluate, pop_size=10, ngen=2)
    """
    size = n_modes
    creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
    creator.create("Individual", list, fitness=creator.FitnessMin)

    toolbox = base.Toolbox()
    toolbox.register("indices", random.sample, range(size), size)
    toolbox.register(
        "individual", tools.initIterate, creator.Individual, toolbox.indices
    )
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("mate", tools.cxOrdered)
    toolbox.register("mutate", tools.mutShuffleIndexes, indpb=0.1)
    toolbox.register("select", tools.selTournament, tournsize=3)

    toolbox.register("evaluate", evaluate)

    def initIndividual(icls, content):
        return icls(content)

    def initPopulation(pcls, ind_init, first_gen):
        return pcls(ind_init(c) for c in first_gen)

    toolbox.register("individual_guess", initIndividual, creator.Individual)

    first_gen = toolbox.population(n=pop_size)[1:]
    first_gen.append([*range(size)])
    toolbox.register(
        "population_guess", initPopulation, list, toolbox.individual_guess, first_gen
    )
    population = toolbox.population_guess()

    toolbox.register("evaluate", evaluate)
    stats = tools.Statistics(key=lambda ind: ind.fitness.values)
    stats.register("avg", np.mean)
    stats.register("std", np.std)
    stats.register("min", np.min)
    stats.register("max", np.max)

    hof = tools.HallOfFame(1)
    pop, logbook = algorithms.eaMuPlusLambda(
        population,
        toolbox,
        mu=pop_size // 2,
        lambda_=pop_size // 2,
        cxpb=0.02,
        mutpb=0.6,
        ngen=ngen,
        halloffame=hof,
        stats=stats,
        verbose=False,
    )
    return np.array(hof[0]), logbook



"""Tests for Hamiltonan Functions."""
from ferrmion import TernaryTree
from ferrmion.hamiltonians import (
    molecular_hamiltonian,
    FermionHamiltonian,
    hubbard_hamiltonian
)
from ferrmion.core import encode_standard
import pytest
import numpy as np
from openfermion import QubitOperator, get_sparse_operator
from scipy.sparse.linalg import eigsh
from pytest import fixture
import logging
from hypothesis import given, settings, strategies as st
from .conftest import diagonalise_pauli_hamiltonian

rng = np.random.default_rng(46392034)
logger = logging.getLogger(__name__)

def test_molecular_hamiltonian_equivalent_explicit_fermion_hamiltonian(n_modes: int = 5):
    print(n_modes)
    ones = rng.random((n_modes,n_modes))
    twos = rng.random((n_modes,n_modes,n_modes,n_modes))
    constant_energy = rng.random()

    molh = molecular_hamiltonian(one_e_coeffs=ones, two_e_coeffs=twos, constant_energy=constant_energy)

    assert ones.shape == (n_modes,n_modes)
    assert twos.shape == (n_modes,n_modes, n_modes,n_modes)
    explicit_molh = FermionHamiltonian()
    explicit_molh.creation().annihilation().with_coefficients(ones)
    explicit_molh.creation().creation().annihilation().annihilation().with_coefficients(twos)
    # explicit_molh.add_constant(constant_energy)

    assert molh._terms.keys() == explicit_molh._terms.keys()
    assert np.all(molh._terms["+-"] == explicit_molh._terms["+-"])
    assert np.all(molh._terms["++--"] == explicit_molh._terms["++--"])

@pytest.mark.parametrize("encoding", ["JW", "BK", "PE", "JKMN"])
def test_encode_standard_water_eigvals_equal_expected(encoding, water_data):
    ones = water_data["ones"]
    twos = water_data["twos"]
    e_nuc = water_data["constant_energy"]

    qham = encode_standard(encoding, 14,14, ["+-","++--"], [ones, twos], e_nuc)
    assert np.isclose(qham["I"*14], -46.465600781952176)

    diag = diagonalise_pauli_hamiltonian(qham, 2)

    assert np.allclose(np.sort(diag), np.sort(water_data["eigvals"])[:2])

@pytest.mark.parametrize("encoding", ["JW", "BK", "PE", "JKMN"])
def test_encode_standard_h2_eigvals_equal_expected(encoding, h2_mol_data_sets):
    ones = h2_mol_data_sets["ones"]
    twos = h2_mol_data_sets["twos"]
    e_nuc = h2_mol_data_sets["constant_energy"]
    n_modes = ones.shape[0]
    qham = encode_standard(encoding, n_modes, n_modes, ["+-","++--"], [ones, twos], e_nuc)

    diag = diagonalise_pauli_hamiltonian(qham, 2*n_modes)
    assert np.allclose(np.sort(diag), np.sort(h2_mol_data_sets["eigvals"]))


@given(n_modes=st.integers(2,5))
@settings(max_examples=10, deadline=None)
def test_encode_standard_eigenalues_constant(n_modes):
    ones = rng.random((n_modes,n_modes))
    twos = rng.random((n_modes,n_modes,n_modes,n_modes))
    e_nuc = rng.random()

    diags = []
    for encoding in ["JW", "BK", "PE", "JKMN"]:
        qham = encode_standard(encoding, n_modes, n_modes, ["+-","++--"], [ones, twos], e_nuc)
        diags.append(diagonalise_pauli_hamiltonian(qham, n_modes))

    for d in diags[1:]:
        assert np.allclose(np.sort(diags[0]), np.sort(d))



"""Init for encodings."""

from .base import FermionQubitEncoding, MajoranaStringEncoding
from .maxnto import MaxNTO
from .standard import (
    bravyi_kitaev,
    bravyi_kitaev_annealed,
    bravyi_kitaev_topphatt,
    jkmn,
    jkmn_annealed,
    jkmn_topphatt,
    jordan_wigner,
    jordan_wigner_annealed,
    jordan_wigner_topphatt,
    parity,
    parity_annealed,
    parity_topphatt,
)
from .ternary_tree import (
    TernaryTree,
)

__all__ = [
    "FermionQubitEncoding",
    "MajoranaStringEncoding",
    "TernaryTree",
    "MaxNTO",
    "jordan_wigner",
    "jordan_wigner_annealed",
    "jordan_wigner_topphatt",
    "parity",
    "parity_annealed",
    "parity_topphatt",
    "bravyi_kitaev",
    "bravyi_kitaev_annealed",
    "bravyi_kitaev_topphatt",
    "jkmn",
    "jkmn_annealed",
    "jkmn_topphatt",
]



"""Ternary Tree fermion to qubit mappings."""

import logging
from copy import deepcopy
from typing import Callable

import numpy as np
from numpy.typing import NDArray

from ferrmion import core
from ferrmion.hamiltonians import FermionHamiltonian, QubitHamiltonian

from .base import FermionQubitEncoding
from .ternary_tree_node import TTNode, node_sorter

logger = logging.getLogger(__name__)

"""Instructions to build a TernaryTree.

The first item in the list gives the qubit-index of the  root node,
followed by the qubit indices of its child nodes (ordered X, Y, Z).
If no child exists, None is used.
"""
type TTFlatpack = list[tuple[int, tuple[int | None, int | None, int | None]]]


class TernaryTree(FermionQubitEncoding):
    """Ternary tree encoding for fermionic operators.

    Attributes:
        n_modes (int): The number of fermionic modes to be encoded.
        n_qubits (int): The number of qubits in encoded operators.
        root (TTNode): The root node of the tree.
        enumeration_scheme (dict[str, tuple[int, int]] | None): The enumeration scheme.

    Methods:
        default_mode_op_map(): Create a default mode operator map for the tree.
        default_enumeration_scheme(): Create a default enumeration scheme for the tree.
        as_dict(): Return the tree structure as a dictionary.
        add_node(node_string: str): Add a node to the tree.
        branch_pauli_map(): Create a map from each branch string to a Pauli string.
        string_pairs(): Return the pair of branch strings which correspond to each node.
        _build_symplectic_matrix(): Build the symplectic matrix for the tree.

    Simple Example:
        >>> from ferrmion.encode.ternary_tree import TernaryTree
        >>> tree = TernaryTree(4)
        >>> tree.add_node('x')
        >>> tree.enumeration_scheme = tree.default_enumeration_scheme()
        >>> tree.as_dict()

    Advanced Usage:
        >>> from ferrmion.encode.ternary_tree import TernaryTree
        >>> tree = TernaryTree(4)
        >>> jordan_wigner = tree.JW()
        >>> bravyi_kitaev = tree.BK()
        >>> parity = tree.Parity()
        >>> minimum_height_tree = tree.JKMN()
    """

    def __init__(
        self,
        n_modes: int,
        n_qubits: None | int = None,
        root_node: TTNode = TTNode(),
    ):
        """Initialise a ternary tree.

        Args:
            n_modes (int): How many fermionic modes in the encoding.
            n_qubits (int): Optional overwrite of number of qubits in target encoding.
            root_node (TTNode): The root node of the tree.
        """
        self.n_modes = n_modes
        self.n_qubits = n_modes if n_qubits is None else n_qubits
        self.root_node = root_node

        if None not in root_node.child_qubit_labels.values():
            self.enumeration_scheme: dict[str, tuple[int, int]] = {
                node: (mode, qubit)
                for mode, (node, qubit) in enumerate(
                    root_node.child_qubit_labels.items()
                )
            }

        self.vacuum_state = np.array([0] * self.n_qubits, dtype=np.uint8)
        self._enumeration_scheme = {}
        super().__init__(self.n_modes, self.n_qubits)

    @classmethod
    def from_hamiltonian_coefficients(cls, coeffs: tuple) -> "TernaryTree":
        """Create an encoding by passing coefficients.

        Args:
            coeffs (tuple): The electron integrals for some hamiltonian.

        Returns:
            FermionQubitEncoding: An initialised encoding.

        Example:
            >>> import numpy as np
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> coeffs = (np.zeros((4, 4)), np.zeros((4, 4, 4, 4)))
            >>> tree = TernaryTree.from_hamiltonian_coefficients(coeffs)
        """
        if not all([set(coeff.shape) == set(coeffs[0].shape) for coeff in coeffs]):
            logger.error("Coeff axes must be of equal size for all terms.")

        return cls(coeffs[0].shape[0])

    @property
    def enumeration_scheme(self) -> dict[str, tuple[int, int]]:
        """Get the enumeration scheme for the tree.

        Note:
            The tuple is organised as (modes, qubits).

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3).JW()
            >>> tree.enumeration_scheme
            {"": (0,0), "z": (1,1), "zz": (2,2)}
        """
        return self._enumeration_scheme

    @enumeration_scheme.setter
    def enumeration_scheme(self, enumeration_dict: dict[str, tuple[int, int]]):
        """Set the enumeration scheme.

        Args:
            enumeration_dict (dict[str, tuple[int, int]]): An dictionary mapping tree nodes to (mode, qubit) indices
        """
        logger.debug("Setting enumeration scheme.")
        error_string = ""
        if set(self.root_node.child_strings) != set(enumeration_dict.keys()):
            error_string += f"Enumeration scheme {enumeration_dict} must contain all nodes {self.root_node.child_strings}.\n"

        modes = set()
        qubits = set()
        for m, q in enumeration_dict.values():
            logger.debug(f"{m=}{q=}")
            modes.add(m)
            qubits.add(q)
        expected_modes = set(range(self.n_modes))
        if set(modes).symmetric_difference(expected_modes):
            error_string += f"Invalid mode labels {set(modes)} in enumeration scheme ({expected_modes=}).\n"
        if len(set(qubits)) != self.n_modes:
            error_string += f"Expected {self.n_modes} qubit labels, got {len(set(qubits))} in enumeration scheme.\n"

        if error_string != "":
            logger.error(error_string)
            raise ValueError(error_string)

        self.default_mode_op_map = [enum[0] for enum in enumeration_dict.values()]
        self._enumeration_scheme = enumeration_dict

    def encode_topphatt(
        self, fham: FermionHamiltonian, parallelize: bool = True
    ) -> QubitHamiltonian:
        """Encode a Hamiltonian, using TOPP-HATT optimisation."""
        sigs, coeffs = fham.signatures_and_coefficients
        ipow, sym, qham = core.encode_topphatt(
            flatpack=self.flatpack(),
            n_qubits=self.n_qubits,
            signatures=deepcopy(sigs),
            coeffs=deepcopy(coeffs),
            constant_energy=fham.constant_energy,
            parallelize=parallelize,
        )
        self._build_symplectic_matrix: Callable = lambda: (ipow, sym)
        self.default_mode_op_map = [*range(self.n_modes)]

        return qham

    def hartree_fock_state(
        self,
        fermionic_hf_state: NDArray[bool],
        mode_op_map: NDArray[np.uint] | list[int] | None = None,
    ):
        """Find the Hartree-Fock state of a majorana string encoding.

        This function calls to the rust implementatin in `src/lib.rs`.
        It assumes that the vacuum state is a single state vector, though the HF state may not be
        The global phase so that the first component state has 0 phase.

        Args:
            fermionic_hf_state (NDArray[int]): An array of mode occupations.
            mode_op_map (dict[int, int]): An array mapping modes to pairs of majorana strings mode_op_map[i]=j => i -> (2j,2j+1)

        Returns:
            NDArray: The Hartree-Fock ground state in computational basis.
        """
        if mode_op_map is None:
            mode_op_map = self.default_mode_op_map

        if isinstance(mode_op_map, list):
            mode_op_map = np.array(mode_op_map, dtype=np.uint)

        ipow, sym = self._build_symplectic_matrix()

        return core.ternary_tree_hartree_fock_state(
            fermionic_hf_state,
            mode_op_map,
            ipow,
            sym,
        )

    def flatpack(self) -> TTFlatpack:
        """Create a TTFlatpack from the tree, which can be passed to rust functions.

        Returns:
            list[tuple[int, tuple[int,int,int]]]
        """
        flatpack: TTFlatpack = []

        to_flatten: list[TTNode] = [self.root_node]
        while len(to_flatten) > 0:
            node: TTNode = to_flatten.pop(0)
            children: list[int | None] = [None, None, None]
            if isinstance(node.x, TTNode):
                to_flatten.append(node.x)
                children[0] = self.enumeration_scheme[node.x.root_path][1]
            if isinstance(node.y, TTNode):
                to_flatten.append(node.y)
                children[1] = self.enumeration_scheme[node.y.root_path][1]
            if isinstance(node.z, TTNode):
                to_flatten.append(node.z)
                children[2] = self.enumeration_scheme[node.z.root_path][1]

            flatpack.append(
                (
                    int(self.enumeration_scheme[node.root_path][1]),
                    (children[0], children[1], children[2]),
                )
            )

        return flatpack

    @classmethod
    def from_flatpack(cls, flatpack: TTFlatpack) -> "TernaryTree":
        """Construct a TernaryTree from a TTFlatpack.

        Args:
            flatpack: The flatpack representation of the tree.

        Returns:
            A new TernaryTree instance.

        Raises:
            TypeError: If the flatpack is invalid.
        """
        if not flatpack:
            raise ValueError("Flatpack cannot be empty")
        used_qubit_indices = [flatpack[0][0]]
        for item in flatpack:
            if item[0] not in used_qubit_indices:
                raise TypeError("Cannot construct TTFlatpack from disconnected nodes.")
            children = item[1]
            if len(children) != 3:
                raise TypeError("TTFlatpack nodes must have three optional children.")
            for child in children:
                if isinstance(child, int):
                    used_qubit_indices.append(child)
                elif child is None:
                    continue
                else:
                    raise TypeError(
                        "TTFlatpack contains child node which is not int | None."
                    )

        ipow, sym = core.flatpack_symplectic_matrix(flatpack)
        max_id = max(item[0] for item in flatpack)
        nodes = [TTNode() for _ in range(max_id + 1)]
        for qubit_index, children in flatpack:
            node = nodes[qubit_index]
            node.qubit_index = qubit_index
            node.x = nodes[children[0]] if children[0] is not None else None
            node.y = nodes[children[1]] if children[1] is not None else None
            node.z = nodes[children[2]] if children[2] is not None else None
        root = nodes[flatpack[0][0]]
        enumeration_scheme = {}
        mode_counter = [0]

        def assign_modes(node, path):
            node.root_path = path
            enumeration_scheme[path] = (mode_counter[0], node.qubit_index)
            mode_counter[0] += 1
            if node.x:
                assign_modes(node.x, path + "x")
            if node.y:
                assign_modes(node.y, path + "y")
            if node.z:
                assign_modes(node.z, path + "z")

        assign_modes(root, "")
        n_modes = len(enumeration_scheme)
        tree = cls(n_modes=n_modes, root_node=root)
        tree.enumeration_scheme = enumeration_scheme
        tree._build_symplectic_matrix = lambda: (ipow, sym)
        return tree

    def default_enumeration_scheme(self) -> dict[str, tuple[int, int]]:
        """Create a default enumeration scheme for the tree.

        Note:
            The tuple is organised as (modes, qubits).

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3).JW()
            >>> tree.default_enumeration_scheme()
            {"": (0,0), "z": (1,1), "zz": (2,2)}
        """
        logger.debug("Setting default enumeration scheme")
        logger.debug("Child strings %s", self.root_node.child_strings)
        enumeration_scheme = {}
        child_labels = self.root_node.child_qubit_labels
        spare_labels: set[int] = set(range(len(child_labels))).difference(
            child_labels.values()
        )
        for mode, (child, qubit) in enumerate(child_labels.items()):
            if qubit is None:
                qubit = spare_labels.pop()
            enumeration_scheme[child] = (int(mode), int(qubit))
        return enumeration_scheme

    def as_dict(self):
        """Return the tree structure as a dictionary."""
        return self.root_node.as_dict()

    def add_node(self, node_string: str) -> "TernaryTree":
        """Add a node to the tree.

        Args:
            node_string (str): The string representation of the node.

        Returns:
            TernaryTree: The tree with the node added.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3)
            >>> tree.add_node('x')
        """
        logger.debug("Adding node %s to TernaryTree", node_string)
        node_string = node_string.lower()
        valid_string = np.all([char in ["x", "y", "z"] for char in node_string])
        if not valid_string:
            raise ValueError("Branch string can only contain x,y,z")

        node = self.root_node
        for char in node_string:
            if isinstance(getattr(node, char), TTNode):
                node = getattr(node, char)
            else:
                node = node.add_child(
                    which_child=char,
                )
        return self

    @property
    def branch_pauli_map(self) -> dict[str, str]:
        """Create a map from each branch string to a Pauli string.

        Returns:
            dict[str, str]: A dictionary of all branch strings with their corresponding Pauli strings.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3)
            >>> tree.add_node('x')
            >>> tree.add_node('xz')
            >>> tree.branch_pauli_map
            {'xx': 'XXI',
            'xzx': 'XZX',
            'y': 'YII',
            'xy': 'XYI',
            'xzz': 'XZZ',
            'xzy': 'XZY',
            'z': 'ZII'}
        """
        logger.debug("Building branch operator map for TernaryTree.")

        branches = self.root_node.branch_strings

        qubit_index = {
            node: qubit for node, (_, qubit) in self.enumeration_scheme.items()
        }
        branch_pauli_map = {}
        for branch in branches:
            branch_pauli_map[branch] = ["I"] * self.n_qubits
            node = self.root_node
            for char in branch:
                node_index = qubit_index[node.root_path]
                branch_pauli_map[branch][node_index] = char.upper()
                node = getattr(node, char, None)

            branch_pauli_map[branch] = "".join(branch_pauli_map[branch])
        logger.debug("Branch pauli map complete")
        logger.debug(branch_pauli_map)
        return branch_pauli_map

    @property
    def string_pairs(self) -> dict[str | int, tuple[str, str]]:
        """Return the pair of branch strings which correspond to each node.

        Returns:
            dict[str, tuple(str,str)]: A dictionary of all node labels, j,  with branch strings (2j, 2j+1).

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3)
            >>> tree.add_node('x')
            >>> tree.add_node('xz')
            >>> tree.string_pairs
            {'': ('xzz', 'y'), 'x': ('xx', 'xy'), 'xz': ('xzx', 'xzy')}
        """
        logger.debug("Building string pairs for TernaryTree.")
        node_set = self.root_node.child_strings

        pairs = {}
        for node_string in node_set:
            node = self.root_node
            for char in node_string:
                node = getattr(node, char)

            x_string = node_string + "x"
            y_string = node_string + "y"
            while x_string in node_set:
                x_string += "z"

            while y_string in node_set:
                y_string += "z"

            if x_string.count("y") % 2 == 0:
                pairs[node.root_path] = x_string, y_string
            elif y_string.count("y") % 2 == 0:
                pairs[node.root_path] = y_string, x_string

        return pairs

    def _build_symplectic_matrix(
        self,
    ) -> tuple[NDArray[np.uint8], NDArray[bool]]:
        """Build the symplectic matrix for the tree.

        Returns:
            NDArray[np.uint8]: Powers of i for each row of the symplectic matrix.
            NDArray[np.uint8]: Symplectic matrix.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3)
            >>> tree.add_node('x')
            >>> tree.enumeration_scheme = tree.default_enumeration_scheme()
            >>> tree._build_symplectic_matrix()
            (array([0, 1, 0, 1, 0, 1], dtype=uint8),
            array([[ True, False, False, False,  True,  True],
                    [ True, False, False,  True, False, False],
                    [ True,  True, False, False, False, False],
                    [ True,  True, False, False,  True, False],
                    [ True, False,  True, False,  True, False],
                    [ True, False,  True, False,  True,  True]]))
        """
        ipow, sym = core.flatpack_symplectic_matrix(self.flatpack())
        reordering_index = np.kron(
            self.default_mode_op_map, np.array([2, 2], dtype=np.uint)
        )
        reordering_index += np.arange(reordering_index.size, dtype=np.uint) % 2
        return ipow[reordering_index], sym[reordering_index, :]

    def JordanWigner(self) -> "TernaryTree":
        """Create a new tree with the Jordan-Wigner encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> jw_tree = TernaryTree(3).JordanWigner()
        """
        return JordanWigner(self.n_modes)

    def JW(self) -> "TernaryTree":
        """Alias for Jordan-Wigner encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> jw_tree = TernaryTree(3).JW()
        """
        return JordanWigner(self.n_modes)

    def ParityEncoding(self) -> "TernaryTree":
        """Create a new tree with the parity encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> parity_tree = TernaryTree(3).ParityEncoding()
        """
        return ParityEncoding(self.n_modes)

    def BravyiKitaev(self) -> "TernaryTree":
        """Create a new tree with the Bravyi-Kitaev encoding.

        Args:
            n_modes (int): The number of fermionic modes.

        Returns:
            TernaryTree: A ternary tree encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> bk_tree = TernaryTree(3).BravyiKitaev()
        """
        return BravyiKitaev(self.n_modes)

    def BK(self) -> "TernaryTree":
        """Alias for Bravyi-Kitaev encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> bk_tree = TernaryTree(3).BK()
        """
        return BravyiKitaev(self.n_modes)

    def JKMN(self) -> "TernaryTree":
        """Create a new tree with the JKMN encoding.

        The JKMN encoding gives a ternary tree with the minimum Pauli-weight.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> min_height_tree = TernaryTree(3).JKMN()
        """
        return JKMN(self.n_modes)


def string_pairing_algorithm(tree: TernaryTree):
    """String-pairing algoritm.

    This is used to produce a map from branches to majorana-indices
    of the root node.

    Args:
        tree (TernaryTree): A Ternary-tree encoding.

    Returns:
        dict[str, int]: A map from branches to majorana mdoe indices.
    """
    logger.debug("Running the string-pairing algorithm.")
    node_set = tree.root_node.child_strings

    branch_majorana_map = {}
    for node_string in node_set:
        # We want to set the majorana indices according to the
        # fermionic ones so that f_i -> (m_2i, m_2i+1)
        fermion_mode = tree.enumeration_scheme[node_string][0]

        x_string = node_string + "x"
        y_string = node_string + "y"
        while x_string in node_set:
            x_string += "z"

        while y_string in node_set:
            y_string += "z"

        if x_string.count("y") % 2 == 0:
            branch_majorana_map[x_string] = 2 * fermion_mode
            branch_majorana_map[y_string] = 2 * fermion_mode + 1
        elif y_string.count("y") % 2 == 0:
            branch_majorana_map[y_string] = 2 * fermion_mode
            branch_majorana_map[x_string] = 2 * fermion_mode + 1

    # We'll place the all-z string after all the required majorana modes
    all_z = "z"
    while all_z in node_set:
        all_z += "z"
    branch_majorana_map[all_z] = 2 * len(node_set) + 1

    logger.debug("String-paring algoithm complete.")
    logger.debug(f"{branch_majorana_map=}")
    return branch_majorana_map


def JordanWigner(n_modes: int) -> TernaryTree:
    """Create a new tree with the Jordan-Wigner encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import JordanWigner
        >>> jw_tree = JordanWigner(3)
    """
    logger.debug("Creating Jordan-Wigner encoding tree")
    new_tree = TernaryTree(
        n_modes=n_modes,
        root_node=TTNode(),
    )
    new_tree.add_node("z" * (n_modes - 1))
    new_tree.enumeration_scheme = new_tree.default_enumeration_scheme()
    return new_tree


def JW(n_modes: int) -> TernaryTree:
    """Alias for Jordan-Wigner encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import JW
        >>> jw_tree = JW(3)
    """
    return JordanWigner(n_modes)


def ParityEncoding(n_modes: int) -> TernaryTree:
    """Create a new tree with the parity encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import ParityEncoding
        >>> parity_tree = ParityEncoding(3)
    """
    logger.debug("Creating parity encoding tree")
    new_tree = TernaryTree(
        n_modes=n_modes,
        root_node=TTNode(),
    )
    new_tree.add_node("x" * (n_modes - 1))
    new_tree.enumeration_scheme = new_tree.default_enumeration_scheme()
    return new_tree


def PE(n_modes: int) -> TernaryTree:
    """Alias for ParityEncoding."""
    return ParityEncoding(n_modes)


def BravyiKitaev(n_modes: int) -> "TernaryTree":
    """Create a new tree with the Bravyi-Kitaev encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import BravyiKitaev
        >>> bk_tree = BravyiKitaev(3)
    """
    logger.debug("Creating Bravyi-Kitaev encoding tree")
    new_tree = TernaryTree(
        n_modes=n_modes,
        root_node=TTNode(),
    )
    branches = ["x"]
    # one is used for root, which is defined
    remaining_qubits = n_modes - 1
    while remaining_qubits > 0:
        new_branches = set()
        for item in branches:
            if remaining_qubits > 0:
                new_tree.add_node(item)
                remaining_qubits -= 1
            else:
                break

            new_branches.add(item + "x")
            new_branches.add(item + "z")
        branches = sorted(list(new_branches), key=node_sorter)
    new_tree.enumeration_scheme = new_tree.default_enumeration_scheme()
    return new_tree


def BK(n_modes: int) -> TernaryTree:
    """Alias for Bravyi-Kitaev encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import BK
        >>> bk_tree = BK(3)
    """
    return BravyiKitaev(n_modes)


def JKMN(n_modes: int) -> TernaryTree:
    """Create a new tree with the JKMN encoding.

    The JKMN encoding gives a ternary tree with the minimum Pauli-weight.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import JKMN
        >>> min_height_tree = JKMN(3)
    """
    logger.debug("Creating JKMN encoding tree.")
    new_tree = TernaryTree(
        n_modes=n_modes,
        root_node=TTNode(),
    )
    branches = ["x", "y", "z"]
    # one is used for root which is defined
    remaining_qubits = n_modes - 1
    while remaining_qubits > 0:
        new_branches = set()
        for item in branches:
            if remaining_qubits > 0:
                new_tree.add_node(item)
                remaining_qubits -= 1
            else:
                break

            new_branches.add(item + "x")
            new_branches.add(item + "y")
            new_branches.add(item + "z")
        branches = sorted(list(new_branches), key=node_sorter)
    new_tree.enumeration_scheme = new_tree.default_enumeration_scheme()
    return new_tree



"""Init for enumeration optimizations."""

import numpy as np

from ferrmion.core import anneal_enumerations as core_anneal_enumerations

from .evolutionary import lambda_plus_mu


def anneal_enemerations(
    template: dict,
    one_e_coeffs: np.ndarray,
    two_e_coeffs: np.ndarray,
    temperature: float | None = None,
    initial_guess=np.typing.ArrayLike | None,
    coefficient_weighted: bool = False,
):
    n_modes = one_e_coeffs.shape[0]
    if temperature is None:
        temperature = float(n_modes)
    if initial_guess is None:
        initial_guess = np.array([*range(n_modes)], dtype=np.uint)
    return core_anneal_enumerations(
        template=template,
        one_e_coeffs=one_e_coeffs,
        two_e_coeffs=two_e_coeffs,
        temperature=temperature,
        initial_guess=initial_guess,
        coefficient_weighted=coefficient_weighted,
    )


__all__ = ["lambda_plus_mu", "anneal_enumerations"]



"""Code to Geneate Hamiltonian Adaptive Ternary Tree from Majorana Hamiltonian."""

import logging
from itertools import permutations
from typing import Iterable

import numpy as np

from ferrmion.encode import TernaryTree
from ferrmion.encode.ternary_tree_node import TTNode

logger = logging.getLogger(__name__)


def _qubit_term_weight(term: Iterable, comb: tuple[int, int, int]) -> int:
    """Find the single-qubit Pauli-weight of majorana terms.

    If any pauli term is found an even number f times, we obtain I, weight = 0.
    If we find all three pauli terms, return I (with an imaginary ccoefficient), weight = 0
    If we find either one pauli or two then the weight = 1.

    Args:
        term (Iterable): Indices of term in our majorana-hamiltonian.
        comb (tuple[int, int, int]): Combination of indices to weigh (x,y,z).

    Returns:
        int: Weight of the term.
    """
    odd_parity_paulis = [sum([t == c for t in term]) % 2 for c in comb]
    non_commuting = sum(odd_parity_paulis) % 3
    return int(non_commuting != 0)


def _reduce_hamiltonian(
    majorana_ham: dict[Iterable[int], float],
    parent_index: int,
    selection: list[int, int, int],
) -> dict[tuple[int, ...], float]:
    """Simplify the Hamiltonian.

    As we increase the qubit number, we iteratively remove majoranas
    which act trivially on the remaining qubits.
    We replace them with the index of their parent string
    as going forward they are identical to the parent string.

    Args:
        majorana_ham (dict[tuple[int,...],float]): Current Hamiltonian.
        parent_index (int): Qubit index of the parent node.
        selection (tuple[int, int, int]): Indices of the majoranas to be replaced.

    Returns:
        dict[tuple[int,...],float]: Reduced Hamiltonian.
    """
    new_ham = {}
    for term, coeff in majorana_ham.items():
        # new_term = tuple(i if i not in selection else parent_index for i in term)
        new_term = tuple(i for i in term if i not in selection) + tuple(
            parent_index for i in term if i in selection
        )
        if len(set(new_term)) >= 1:
            new_ham[new_term] = new_ham.get(new_term, 0) + coeff
    return new_ham


def hamiltonian_adaptive_ternary_tree(
    majorana_ham: dict[Iterable[int], float], n_modes: int
) -> TernaryTree:
    """Construct an adaptive ternary tree from a majorana Hamiltonian.

    Args:
        majorana_ham (dict[tuple[int,...],float]): Majorana Hamiltonian to encode.
        n_modes (int): Number of fermionic modes in the system.

    Returns:
        TTNode: Root node of the constructed ternary tree.
    """
    # We need 2*M +1 leaves and M nodes.
    nodes: dict[int, TTNode | None] = {i: None for i in range(2 * n_modes + 1)}
    for i in range(n_modes):
        nodes[2 * n_modes + 1 + i] = TTNode(qubit_label=i)

    # Start with all the leaves unassigned
    unassigned = {*range(2 * n_modes + 1)}

    # We create two maps, of z_ancestors and z_descendants
    ancestor_map = {i: i for i in nodes}
    descendant_map = {i: i for i in nodes}

    total_weight = 0
    for i in range(n_modes):
        parent_index = 2 * n_modes + 1 + i
        parent = nodes[parent_index]

        min = np.inf
        for comb in permutations(unassigned, 2):
            small_y = None
            small_x = None
            # This way x index will be higher term - more often node.
            # z_index, x_index= comb
            x_index, z_index = comb
            small_x = descendant_map[x_index]

            # discard this combination
            if small_x == 2 * n_modes:
                continue

            if small_x % 2 == 0:
                small_y = small_x + 1
            else:
                small_y = small_x - 1
            # We can't use this index for y a
            # it has been used in the combination already
            # so we'd be replacing our x or z!
            if small_y in comb:
                continue

            y_index = ancestor_map[small_y]

            if y_index in comb:
                continue

            if small_x % 2 == 0:
                comb = np.array([x_index, y_index, z_index], dtype=np.uint)
            else:
                comb = np.array([y_index, x_index, z_index], dtype=np.uint)
            comb = [int(i) for i in comb]
            weight = np.sum(
                [_qubit_term_weight(term, comb) for term in majorana_ham.keys()]
            )
            if weight < min:
                min = weight
                selection = comb
            # would be better to break on zero
            # if weight == 0:
            #     break

        total_weight += min
        # Now find the Y pair of the x-node
        for i, char in zip(selection, ["x", "y", "z"]):
            if i in unassigned:
                unassigned.remove(i)

            if isinstance(nodes.get(i, None), TTNode):
                parent.add_child(which_child=char, child_node=nodes.get(i))
            else:
                parent.leaf_majorana_indices[char] = i

        z_index = selection[2]
        z_desc = descendant_map[z_index]
        descendant_map[parent_index] = z_desc
        ancestor_map[z_index] = parent_index
        ancestor_map[z_desc] = parent_index

        unassigned.add(parent_index)

        majorana_ham = _reduce_hamiltonian(majorana_ham, parent_index, selection)

    if len(unassigned) != 1:
        raise ValueError("Not all nodes assigned by HATT.")

    last_node = nodes[unassigned.pop()]
    if isinstance(last_node, TTNode):
        root = last_node
    else:
        raise ValueError("Hatt root node is not a TTNode object.")

    tree = TernaryTree(n_modes=n_modes, root_node=root)
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    tree.pauli_weight = total_weight
    return tree


def fast_hatt(
    majorana_ham: dict[Iterable[int], float],
    n_modes: int,
    coeff_weight: bool = False,
) -> TernaryTree:
    """Construct an adaptive ternary tree from a majorana Hamiltonian.

    Args:
        majorana_ham (dict[tuple[int,...],float]): Majorana Hamiltonian to encode.
        n_modes (int): Number of fermionic modes in the system.
        coeff_weight (bool): Multiply Pauli-weight by coefficient norm.

    Returns:
        TTNode: Root node of the constructed ternary tree.
    """
    # if coeff_weight:
    # majorana_ham = {k:v for k,v in sorted(majorana_ham.items(), key=lambda item: sum(item[0])*item[1], reverse=True)}
    # else:
    # majorana_ham = {k:v for k,v in sorted(majorana_ham.items(), key=lambda item: sum(item[0]), reverse=True)}

    n_leaves = 2 * n_modes + 1
    # We need 2*M +1 leaves and M nodes.
    nodes: dict[int, TTNode | None] = {i: None for i in range(n_leaves)}
    for i in range(n_modes):
        nodes[n_leaves + i] = TTNode(qubit_label=i)

    # Start with all the leaves unassigned
    unassigned = [*range(n_leaves)]
    unassigned.reverse()

    # We create two maps, of z_ancestors and z_descendants
    ancestor_map = {i: i for i in range(n_leaves + n_modes)}
    descendant_map = {i: i for i in range(n_leaves + n_modes)}

    total_weight = 0
    for i in range(n_modes + 1):
        parent_index = n_leaves + i
        parent = nodes[parent_index]

        min_weight = np.inf
        selection = [None, None, None]
        previous = [None, None, None]
        # reverse because best is usually at the end
        # reversed_combinations = [c for c in combinations(unassigned, 2)]
        if i == 0:
            comb_iterator = ([n_leaves - 1, i] for i in range(n_leaves - 1)[::-1])
        else:
            comb_iterator = permutations(unassigned, 2)

        for comb in comb_iterator:
            small_y = None
            small_x = None
            # This way x index will be higher term - more often node.
            z_index, x_index = comb
            # x_index, z_index = comb

            small_x = descendant_map[x_index]

            # discard this combination
            if small_x == 2 * n_modes:
                # print("small x is all z")
                continue

            if small_x % 2 == 0:
                small_y = small_x + 1
            else:
                small_y = small_x - 1
            # We can't use this index for y a
            # it has been used in the combination already
            # so we'd be replacing our x or z!
            if small_y in comb:
                # print("small_y in comb")
                continue

            y_index = ancestor_map[small_y]

            if y_index in comb:
                # print("y index in comb")
                continue

            if small_x % 2 == 0:
                comb = [x_index, y_index, z_index]
            else:
                comb = [y_index, x_index, z_index]
            # comb = [int(i) for i in comb]

            # W3 can end uop with the same combination in two different ways
            if comb == selection:
                continue
            if comb == previous:
                continue
            previous = comb
            # tc = tuple(comb)
            # if tc in checked:
            #     continue
            # checked.add(tc)

            weight = 0
            if coeff_weight:
                for key, val in majorana_ham.items():
                    if min(comb) > key[-1]:
                        continue
                    elif max(comb) < key[0]:
                        continue
                    else:
                        odd_parity_paulis = [
                            sum([t == c for t in key]) % 2 for c in comb
                        ]
                        non_commuting = sum(odd_parity_paulis) % 3
                        weight += int(non_commuting != 0)
                        weight *= abs(val)
                    if weight > min_weight:
                        break
            else:
                for key in majorana_ham.keys():
                    if min(comb) > key[-1]:
                        continue
                    elif max(comb) < key[0]:
                        continue
                    else:
                        odd_parity_paulis = [
                            sum([t == c for t in key]) % 2 for c in comb
                        ]
                        non_commuting = sum(odd_parity_paulis) % 3
                        weight += int(non_commuting != 0)
                        if weight > min_weight:
                            break
                # weight = np.sum(
                #     [_qubit_term_weight(term, comb) for term in majorana_ham.keys()]
                # )
            if weight < min_weight:
                # print(f"NEW Min Node:{i}, Parent Index: {parent_index}, Comb: {comb}, Old Min:{min_weight }, New Min:{weight}")
                min_weight = weight
                selection = comb
            elif weight == min_weight:
                #     # print(f"SAME Min Node:{i}, Parent Index: {parent_index}, Comb: {comb}, Old Min:{min_weight }, New Min:{weight}")
                min_weight = weight
                selection = comb
            # would be better to break on zero
            # if weight == 0:
            #     break

        total_weight += min_weight
        # print(f"{selection=}")
        # Now find the Y pair of the x-node
        unassigned = [u for u in unassigned if u not in selection]
        for child_index, char in zip(selection, ["x", "y", "z"]):
            if isinstance(nodes.get(child_index, None), TTNode):
                # print(f"{child_index} {nodes[child_index]}")
                # print(f"Child node {char=} with index {child_index=} is node {nodes.get(child_index).qubit_label=}")
                parent.add_child(which_child=char, child_node=nodes.get(child_index))
            else:
                parent.leaf_majorana_indices[char] = child_index

        # unassigned.append(parent_index)
        unassigned = [parent_index] + unassigned

        if i + 1 == n_modes:
            break

        z_index = selection[2]
        z_desc = descendant_map[z_index]
        descendant_map[parent_index] = z_desc
        ancestor_map[z_index] = parent_index
        ancestor_map[z_desc] = parent_index

        # print("START reducing hamiltonian")
        majorana_ham = _reduce_hamiltonian(majorana_ham, parent_index, selection)
        # print("STOP reducing hamiltonian")

    if len(unassigned) != 1:
        raise ValueError(f"Not all nodes assigned by HATT. {unassigned=}")

    last_node = nodes[unassigned[0]]
    if isinstance(last_node, TTNode):
        root = last_node
    else:
        raise ValueError("Hatt root node is not a TTNode object.")

    tree = TernaryTree(n_modes=n_modes, root_node=root)
    tree.enumeration_scheme = tree.default_enumeration_scheme()
    tree.pauli_weight = total_weight
    print("Total Weight: ", total_weight)
    return tree
